# -*- coding: utf-8 -*-
#
from functools import lru_cache
from warnings import warn

import pkg_resources

from pkg_resources import DistributionNotFound, UnknownExtra


# In dev mode, the extras-loader doesn't check if extension is installed before loading it.
# This is useful when testing, when we're working on a codebase of a library that is not yet installed.
dev_mode = False


@lru_cache()
def check_extra_installed(extra_name: str):
    if dev_mode:
        return True
    check_name = "pyshacl[" + extra_name + "]"
    # first check if pyshacl is installed using the normal means
    try:
        _ = pkg_resources.require("pyshacl")
    except DistributionNotFound:
        # Hmm, it thinks pyshacl isn't installed. Can't even check for extras
        return None
    try:
        _ = pkg_resources.require(check_name)
        return True
    except UnknownExtra:
        # That extra doesn't exist in this version of pyshacl
        warn(Warning("Extra \"{}\" doesn't exist in this version of pyshacl.".format(extra_name)))
        return False
    except DistributionNotFound:
        # That extra is not installed right now
        return False
    except BaseException:
        raise



import typing

from typing import List, Sequence, Type, Union
from warnings import warn

from .constraints import ConstraintComponent
from .consts import SH, RDF_type, RDFS_subClassOf, SH_parameter, SH_select, SH_SPARQLTargetType
from .errors import ConstraintLoadError, ShapeLoadError
from .helper import get_query_helper_cls
from .parameter import SHACLParameter
from .pytypes import GraphLike


if typing.TYPE_CHECKING:
    from .shapes_graph import ShapesGraph

SH_labelTempalate = SH.labelTemplate
SH_Target = SH.Target
SH_TargetType = SH.TargetType
SH_JSTarget = SH.JSTarget
SH_JSTargetType = SH.JSTargetType
SH_SPARQLTarget = SH.SPARQLTarget


class SHACLTargetType(object):
    __slots__ = ("sg", "node", "parameters", "label_template")

    def __init__(self, t_node, sg):
        """

        :param t_node:
        :type t_node: rdflib.Identifier
        :param sg:
        :type sg: ShapesGraph
        """
        super(SHACLTargetType, self).__init__()
        self.node = t_node
        self.sg = sg
        params = list(sg.objects(t_node, SH_parameter))
        self.parameters = [SHACLParameter(sg, p) for p in params]  # type: List[SHACLParameter]
        ltemps = list(sg.objects(t_node, SH_labelTempalate))
        if len(ltemps) < 1:
            self.label_template = None
        elif len(ltemps) > 1:
            raise ConstraintLoadError(
                "SHACLTargetType cannot have more than one value for sh:labelTemplate.",
                "https://www.w3.org/TR/shacl-af/#SPARQLTargetType",
            )
        else:
            self.label_template = ltemps[0]

    def apply(self):
        self.sg.add_shacl_target_type(self.node, self)

    def check_params(self, target_declaration):
        assert False  # is this even used?
        param_kv = {}
        for p in self.parameters:
            n = p.node
            vals = set(self.sg.objects(target_declaration, n))
            if len(vals) < 1:
                if p.optional:
                    continue
                raise ShapeLoadError(
                    "sh:target does not have a value for {}".format(n),
                    "https://www.w3.org/TR/shacl-af/#SPARQLTargetType",
                )
            if len(vals) > 1:
                warn(Warning("Found more than one value for {} on sh:target. Using just first one.".format(n)))
            param_kv[p] = next(iter(vals))
        return param_kv

    def bind(self, shape, target_declaration):
        assert False  # is this even used?
        param_vals = self.check_params(target_declaration)
        return BoundSHACLTargetType(self, target_declaration, shape, param_vals)


class TargetDeclarationWrapper(object):
    __slots__ = ("sg", "node")

    def __init__(self, sg, node):
        self.sg = sg
        self.node = node

    def objects(self, pred):
        return self.sg.objects(self.node, pred)


class BoundSHACLTargetType(ConstraintComponent):
    __slots__ = ('target_type', 'target_declaration', 'param_vals')

    def __init__(self, target_type, target_declaration, shape, param_vals=None):
        """

        :param target_type: The source TargetType, this is needed to bind the parameters in the query_helper
        :type target_type: SPARQLConstraintComponent
        :param shape:
        :type shape: pyshacl.shape.Shape
        :param param_vals:
        :type param_vals: Dict[SHACLParameter, Any]
        """
        super(BoundSHACLTargetType, self).__init__(shape)
        self.target_type = target_type
        sg = self.shape.sg
        self.target_declaration = TargetDeclarationWrapper(sg, target_declaration)
        self.param_vals = param_vals

    @classmethod
    def constraint_parameters(cls):
        return []

    @classmethod
    def constraint_name(cls):
        return "SPARQLTargetType"

    @classmethod
    def shacl_constraint_class(cls):
        return SH_SPARQLTargetType

    def evaluate(self, target_graph: GraphLike, focus_value_nodes: typing.Dict, _evaluation_path: List):
        raise NotImplementedError()

    def find_targets(self, data_graph):
        return NotImplementedError()


class BoundSPARQLTargetType(BoundSHACLTargetType):
    __slots__ = ('query_helper',)

    def __init__(self, target_type, target_declaration, shape):
        super(BoundSPARQLTargetType, self).__init__(target_type, target_declaration, shape)
        params = self.target_type.parameters
        SPARQLQueryHelper = get_query_helper_cls()
        self.query_helper = SPARQLQueryHelper(
            self.target_declaration, self.target_type.node, self.target_type.select, params
        )
        # Setting self.shape into QueryHelper automatically applies query_helper.bind_params and bind_messages
        self.query_helper.collect_prefixes()

    def find_targets(self, data_graph):
        qh = self.query_helper
        bind_vals = qh.param_bind_map
        # Don't pre-bind variables here!
        # init_binds, sparql_text = qh.pre_bind_variables(self.target_type.node, extravars=bind_vals.keys())
        # init_binds.update(bind_vals)
        sparql_text = qh.apply_prefixes(qh.select_text)
        results = data_graph.query(sparql_text, initBindings=bind_vals)
        return results


class SPARQLTargetType(SHACLTargetType):
    __slots__ = ('select',)

    def __init__(self, t_node, sg):
        super(SPARQLTargetType, self).__init__(t_node, sg)
        selects = list(self.sg.objects(self.node, SH_select))
        num_selects = len(selects)
        if num_selects > 1:
            raise ConstraintLoadError(
                "SPARQLTargetType cannot have more than one value for sh:select.",
                "https://www.w3.org/TR/shacl-af/#SPARQLTargetType",
            )
        elif num_selects < 1:
            raise ConstraintLoadError(
                "SPARQLTargetType must have a value for sh:select.",
                "https://www.w3.org/TR/shacl-af/#SPARQLTargetType",
            )
        self.select = selects[0]

    def bind(self, shape, target_declaration):
        return BoundSPARQLTargetType(self, target_declaration, shape)


def gather_target_types(shacl_graph: 'ShapesGraph') -> Sequence[Union['SHACLTargetType', 'SPARQLTargetType']]:
    """

    :param shacl_graph:
    :type shacl_graph: ShapesGraph
    :return:
    :rtype: [SHACLTargetType]
    """
    all_target_types: List[Union['SHACLTargetType', 'SPARQLTargetType']] = []
    sub_targets = set(shacl_graph.subjects(RDFS_subClassOf, SH_Target))

    # remove these two which are the known native types in shacl.ttl
    sub_targets = sub_targets.difference({SH_JSTarget, SH_SPARQLTarget})

    if shacl_graph.js_enabled:
        from pyshacl.extras.js.target import JSTargetType

        use_js: Union[bool, Type] = JSTargetType
    else:
        use_js = False

    for s in sub_targets:
        types = set(shacl_graph.objects(s, RDF_type))
        found = False
        if SH_SPARQLTargetType in types:
            all_target_types.append(SPARQLTargetType(s, shacl_graph))
            found = True
        if SH_JSTargetType in types:
            found = True
            if use_js:
                all_target_types.append(JSTargetType(s, shacl_graph))
            else:
                pass  # JS Mode is not enabled. Silently ignore JSTargetTypes
        if not found:
            warn(Warning("The only SHACLTargetType currently implemented is SPARQLTargetType."))

    return all_target_types


def apply_target_types(tts: Sequence):
    for t in tts:
        t.apply()



#
#
import typing

from typing import Dict, List

from rdflib import Literal

from pyshacl.constraints import ConstraintComponent
from pyshacl.consts import SH, SH_js, SH_message
from pyshacl.errors import ConstraintLoadError
from pyshacl.pytypes import GraphLike

from .js_executable import JSExecutable


if typing.TYPE_CHECKING:
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph


SH_JSConstraint = SH.JSConstraint
SH_JSConstraintComponent = SH.JSConstraintComponent


class JSConstraintImpl(JSExecutable):
    __slots__ = ("messages",)

    def __init__(self, shapes_graph: 'ShapesGraph', node):
        super(JSConstraintImpl, self).__init__(shapes_graph, node)
        msgs_iter = shapes_graph.objects(node, SH_message)
        self.messages = []
        for m in msgs_iter:
            if not isinstance(m, Literal):
                raise ConstraintLoadError(
                    "JSConstraint sh:message must be a RDF Literal.",
                    "https://www.w3.org/TR/shacl-js/#js-constraints",
                )
            if not isinstance(m.value, str):
                raise ConstraintLoadError(
                    "JSConstraint sh:message must be a RDF Literal with type string.",
                    "https://www.w3.org/TR/shacl-js/#js-constraints",
                )
            self.messages.append(m)

    def make_messages(self, args_map=None):
        if args_map is None:
            return self.messages
        ret_msgs = []
        for m in self.messages:
            this_m = m.value[:]
            for a, v in args_map.items():
                replace_me = "{$" + str(a) + "}"
                if isinstance(v, Literal):
                    v = v.value
                this_m = this_m.replace(replace_me, str(v))
            ret_msgs.append(Literal(this_m))
        return ret_msgs


class JSConstraint(ConstraintComponent):

    shacl_constraint_component = SH_JSConstraint

    def __init__(self, shape: 'Shape'):
        super(JSConstraint, self).__init__(shape)
        js_decls = list(self.shape.objects(SH_js))
        if len(js_decls) < 1:
            raise ConstraintLoadError(
                "JSConstraint must have at least one sh:js predicate.",
                "https://www.w3.org/TR/shacl-js/#js-constraints",
            )
        self.js_impls = [JSConstraintImpl(shape.sg, j) for j in js_decls]

    @classmethod
    def constraint_parameters(cls):
        return [SH_js]

    @classmethod
    def constraint_name(cls):
        return "JSConstraint"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[Literal]:
        return [Literal("Javascript Function generated constraint validation reports.")]

    def evaluate(self, data_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List):
        """
        :type data_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports = []
        non_conformant = False
        for c in self.js_impls:
            _n, _r = self._evaluate_js_exe(data_graph, focus_value_nodes, c)
            non_conformant = non_conformant or _n
            reports.extend(_r)
        return (not non_conformant), reports

    def _evaluate_js_exe(self, data_graph, f_v_dict, js_impl: JSConstraintImpl):
        reports = []
        non_conformant = False
        for f, value_nodes in f_v_dict.items():
            for v in value_nodes:
                failed = False
                try:
                    args_map = {'this': f, 'value': v}
                    if self.shape.is_property_shape:
                        args_map['path'] = self.shape.path()
                    res_dict = js_impl.execute(data_graph, args_map)
                    result = res_dict['_result']
                    if result is True:
                        continue
                    msgs = js_impl.make_messages(args_map)
                    if isinstance(result, list):
                        pass
                    else:
                        result = [result]
                    for res in result:
                        if isinstance(res, Literal):
                            res = res.value
                        if isinstance(res, bool):
                            if res:
                                continue
                            else:
                                failed = True
                                reports.append(self.make_v_result(data_graph, f, value_node=v, extra_messages=msgs))
                        elif isinstance(res, str):
                            failed = True
                            msgs.append(Literal(res))
                            reports.append(self.make_v_result(data_graph, f, value_node=v, extra_messages=msgs))
                        elif isinstance(res, dict):
                            failed = True
                            args_map2 = args_map.copy()
                            val = res.get('value', None)
                            if val is None:
                                val = v
                            args_map2['value'] = val
                            path = res.get('path', None) if not self.shape.is_property_shape else None
                            if path is not None:
                                args_map2['value'] = path
                            msgs = js_impl.make_messages(args_map2)
                            message = res.get('message', None)
                            if message is not None:
                                msgs.append(Literal(message))
                            reports.append(
                                self.make_v_result(
                                    data_graph, f, value_node=val, result_path=path, extra_messages=msgs
                                )
                            )
                except Exception as e:
                    print(e)
                    raise
                if failed:
                    non_conformant = True
        return non_conformant, reports



# -*- coding: utf-8 -*-
import logging
import warnings

import rdflib

from .constraints.constraint_component import CustomConstraintComponentFactory
from .constraints.core.logical_constraints import SH_and, SH_not, SH_or, SH_xone
from .constraints.core.shape_based_constraints import SH_qualifiedValueShape
from .consts import (
    SH,
    OWL_Class,
    OWL_DatatypeProperty,
    RDF_Property,
    RDF_type,
    RDFS_Class,
    RDFS_subClassOf,
    SH_ConstraintComponent,
    SH_node,
    SH_NodeShape,
    SH_path,
    SH_property,
    SH_PropertyShape,
    SH_targetClass,
    SH_targetNode,
    SH_targetObjectsOf,
    SH_targetSubjectsOf,
)
from .errors import ShapeLoadError
from .shape import Shape


class ShapesGraph(object):
    system_triples = [(OWL_Class, RDFS_subClassOf, RDFS_Class), (OWL_DatatypeProperty, RDFS_subClassOf, RDF_Property)]

    def __init__(self, graph, logger=None):
        """
        ShapesGraph
        :param graph:
        :type graph: rdflib.Graph
        :param logger:
        :type logger: logging.Logger|None
        """
        assert isinstance(graph, (rdflib.Dataset, rdflib.ConjunctiveGraph, rdflib.Graph))
        self.graph = graph
        if isinstance(self.graph, rdflib.Dataset):
            self.graph.default_union = True
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self._node_shape_cache = {}
        self._shapes = None
        self._custom_constraints = None
        self._shacl_functions = {}
        self._shacl_target_types = {}
        self._use_js = False
        self._add_system_triples()

    def enable_js(self):
        self._use_js = True

    @property
    def js_enabled(self):
        return bool(self._use_js)

    def _add_system_triples(self):
        if isinstance(self.graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            g = next(iter(self.graph.contexts()))
        else:
            g = self.graph
        for t in self.system_triples:
            g.add(t)

    def subjects(self, p, o):
        return self.graph.subjects(p, o)

    def predicates(self, s, o):
        return self.graph.predicates(s, o)

    def objects(self, s, p):
        return self.graph.objects(s, p)

    def subject_objects(self, p):
        return self.graph.subject_objects(p)

    def subject_predicates(self, o):
        return self.graph.subject_predicates(o)

    def predicate_objects(self, s):
        return self.graph.predicate_objects(s)

    @property
    def custom_constraints(self):
        if self._custom_constraints is None:
            self._custom_constraints = self._find_custom_constraints()
        return self._custom_constraints

    def _find_custom_constraints(self):
        g = self.graph
        constraint_component_set = set(g.subjects(RDF_type, SH_ConstraintComponent))
        constraint_subclasses = set(g.subjects(RDFS_subClassOf, SH_ConstraintComponent))
        for sc in iter(constraint_subclasses):
            subclass_components = set(g.subjects(RDF_type, sc))
            constraint_component_set.update(subclass_components)
        components = set()
        for c in iter(constraint_component_set):
            if c.startswith(SH):
                # ignore all constraint components from shacl.ttl, these are all hardcoded into PySHACL
                continue
            components.add(CustomConstraintComponentFactory(self, c))
        return components

    def add_shacl_function(self, uri, function, optionals):
        uri = str(uri)
        if uri in self._shacl_functions:
            warnings.warn(Warning("SHACLFunction {} is already registered.".format(uri)))
        else:
            self._shacl_functions[uri] = (function, optionals)

    def get_shacl_function(self, uri):
        uri = str(uri)
        try:
            f = self._shacl_functions[uri]
        except LookupError:
            raise KeyError("SHACLFunction {} not found.".format(uri))
        return f

    def remove_shacl_function(self, uri, function):
        uri = str(uri)
        try:
            f, _ = self.get_shacl_function(uri)
            if f != function:
                warnings.warn(
                    Warning("Cannot remove a different function than what was registered for {}.".format(uri))
                )
                return
        except KeyError:
            return  # Not registered
        del self._shacl_functions[uri]

    def add_shacl_target_type(self, uri, tt):
        uri = str(uri)
        if uri in self._shacl_target_types:
            warnings.warn(Warning("SHACL TargetType {} is already registered.".format(uri)))
        else:
            self._shacl_target_types[uri] = tt

    def get_shacl_target_type(self, uri):
        uri = str(uri)
        try:
            f = self._shacl_target_types[uri]
        except LookupError:
            raise KeyError("SHACL TargetType {} not found.".format(uri))
        return f

    @property
    def shapes(self):
        """

        :returns: [Shape]
        :rtype: list(pyshacl.shape.Shape)
        """
        if len(self._node_shape_cache) < 1:
            self._build_node_shape_cache()
        return self._node_shape_cache.values()

    def lookup_shape_from_node(self, node) -> Shape:
        # This will throw a KeyError if it is not found. This is intentionally not caught here.
        return self._node_shape_cache[node]

    """
    A shape is an IRI or blank node s that fulfills at least one of the following conditions in the shapes graph:

        s is a SHACL instance of sh:NodeShape or sh:PropertyShape.
        s is subject of a triple that has sh:targetClass, sh:targetNode, sh:targetObjectsOf or sh:targetSubjectsOf as predicate.
        s is subject of a triple that has a parameter as predicate.
        s is a value of a shape-expecting, non-list-taking parameter such as sh:node, or a member of a SHACL list that is a value of a shape-expecting and list-taking parameter such as sh:or.
    """

    def _build_node_shape_cache(self):
        """
        :returns: None
        :rtype: NoneType
        """
        g = self.graph
        defined_node_shapes = set(g.subjects(RDF_type, SH_NodeShape))
        for s in defined_node_shapes:
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) > 0:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a NodeShape cannot be the subject of a 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#node-shapes",
                )

        defined_prop_shapes = set(g.subjects(RDF_type, SH_PropertyShape))
        found_prop_shapes_paths = dict()
        for s in defined_prop_shapes:
            if s in defined_node_shapes:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a NodeShape cannot also be defined as a PropertyShape.",
                    "https://www.w3.org/TR/shacl/#node-shapes",
                )
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) < 1:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a PropertyShape must be the subject of a 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            elif len(path_vals) > 1:
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "A shape defined as a PropertyShape cannot have more than one 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            found_prop_shapes_paths[s] = path_vals[0]

        has_target_class = {s for s, o in g.subject_objects(SH_targetClass)}
        has_target_node = {s for s, o in g.subject_objects(SH_targetNode)}
        has_target_objects_of = {s for s, o in g.subject_objects(SH_targetObjectsOf)}
        has_target_subjects_of = {s for s, o in g.subject_objects(SH_targetSubjectsOf)}
        subject_shapes = set(has_target_class).union(
            set(has_target_node).union(set(has_target_objects_of).union(set(has_target_subjects_of)))
        )

        # implicit shapes: their subjects must be shapes
        subject_of_property = {s for s, o in g.subject_objects(SH_property)}
        subject_of_node = {s for s, o in g.subject_objects(SH_node)}
        subject_shapes = subject_shapes.union(set(subject_of_property).union(set(subject_of_node)))

        # shape-expecting properties, their values must be shapes.
        value_of_property = {o for s, o in g.subject_objects(SH_property)}
        value_of_node = {o for s, o in g.subject_objects(SH_node)}
        value_of_not = {o for s, o in g.subject_objects(SH_not)}
        value_of_qvs = {o for s, o in g.subject_objects(SH_qualifiedValueShape)}
        value_of_shape_expecting = set(value_of_property).union(
            set(value_of_node).union(set(value_of_not).union(set(value_of_qvs)))
        )

        value_of_and = {o for s, o in g.subject_objects(SH_and)}
        value_of_or = {o for s, o in g.subject_objects(SH_or)}
        value_of_xone = {o for s, o in g.subject_objects(SH_xone)}
        value_of_s_list_expecting = set(value_of_and).union(set(value_of_or).union(set(value_of_xone)))

        for lst in value_of_s_list_expecting:
            list_contents = set(g.items(lst))
            if len(list_contents) < 1:
                # TODO:coverage: we don't have any tests for invalid shape lists
                raise ShapeLoadError(
                    "A Shape-Expecting & List-Expecting predicate should get a well-formed RDF list with 1 or more members.",
                    "https://www.w3.org/TR/shacl/#shapes-recursion",
                )
            for s in list_contents:
                value_of_shape_expecting.add(s)

        found_node_shapes = set()
        found_prop_shapes = set()
        for s in subject_shapes:
            if s in defined_node_shapes or s in defined_prop_shapes:
                continue
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) < 1:
                found_node_shapes.add(s)
            elif len(path_vals) > 1:
                # TODO:coverage: we don't have any tests for invalid implicit shapes
                raise ShapeLoadError(
                    "An implicit PropertyShape cannot have more than one 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            else:
                # TODO:coverage: we don't have this case where an implicit shape is a property shape.
                found_prop_shapes.add(s)
                found_prop_shapes_paths[s] = path_vals[0]
        for s in value_of_shape_expecting:
            if (
                s in defined_node_shapes
                or s in defined_prop_shapes
                or s in found_prop_shapes
                or s in found_node_shapes
            ):
                continue
            path_vals = list(g.objects(s, SH_path))
            if len(path_vals) < 1:
                found_node_shapes.add(s)
            elif len(path_vals) > 1:
                # TODO:coverage: we don't have any tests for invalid implicit shapes
                raise ShapeLoadError(
                    "An implicit PropertyShape cannot have more than one 'sh:path' predicate.",
                    "https://www.w3.org/TR/shacl/#property-shapes",
                )
            else:
                found_prop_shapes.add(s)
                found_prop_shapes_paths[s] = path_vals[0]
        for node_shape in defined_node_shapes.union(found_node_shapes):
            if node_shape in self._node_shape_cache:
                # TODO:coverage: we don't have any tests where a shape is loaded twice
                raise ShapeLoadError("That shape has already been loaded!", "None")
            s = Shape(self, node_shape, p=False, logger=self.logger)
            self._node_shape_cache[node_shape] = s
        for prop_shape in defined_prop_shapes.union(found_prop_shapes):
            if prop_shape in self._node_shape_cache:
                # TODO:coverage: we don't have any tests where a shape is loaded twice
                raise ShapeLoadError("That shape has already been loaded!", "None")
            prop_shape_path = found_prop_shapes_paths[prop_shape]
            s = Shape(self, prop_shape, p=True, path=prop_shape_path, logger=self.logger)
            self._node_shape_cache[prop_shape] = s



# -*- coding: utf-8 -*-
#
"""A collection of handy utility RDF functions, will one day be split out into its own installable module."""

from .clone import clone_blank_node, clone_graph, clone_literal, clone_node, mix_datasets, mix_graphs  # noqa: F401
from .compare import compare_blank_node, compare_literal, compare_node, order_graph_literal  # noqa: F401
from .load import get_rdf_from_web, load_from_source  # noqa: F401
from .stringify import stringify_blank_node, stringify_graph, stringify_literal, stringify_node  # noqa: F401



# -*- coding: utf-8 -*-
from rdflib.namespace import Namespace


RDF_PFX = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
RDFS_PFX = 'http://www.w3.org/2000/01/rdf-schema#'
OWL_PFX = 'http://www.w3.org/2002/07/owl#'
SH_PFX = 'http://www.w3.org/ns/shacl#'
RDF = Namespace(RDF_PFX)
RDFS = Namespace(RDFS_PFX)
OWL = Namespace(OWL_PFX)
SH = Namespace(SH_PFX)

# Classes
RDF_Property = RDF.Property
RDF_List = RDF.List
RDFS_Resource = RDFS.Resource
RDFS_Class = RDFS.Class
OWL_Ontology = OWL.Ontology
OWL_Class = OWL.Class
OWL_DatatypeProperty = OWL.DatatypeProperty
SH_NodeShape = SH.NodeShape
SH_PropertyShape = SH.PropertyShape
SH_ValidationResult = SH.ValidationResult
SH_ValidationReport = SH.ValidationReport
SH_Violation = SH.Violation
SH_Info = SH.Info
SH_Warning = SH.Warning
SH_IRI = SH.IRI
SH_BlankNode = SH.BlankNode
SH_Literal = SH.Literal
SH_BlankNodeOrIRI = SH.BlankNodeOrIRI
SH_BlankNodeORLiteral = SH.BlankNodeOrLiteral
SH_IRIOrLiteral = SH.IRIOrLiteral
SH_ConstraintComponent = SH.ConstraintComponent
SH_PropertyConstraintComponent = SH.PropertyConstraintComponent
SH_NodeConstraintComponent = SH.NodeConstraintComponent
SH_SHACLFunction = SH.SHACLFunction
SH_SPARQLFunction = SH.SPARQLFunction
SH_SPARQLRule = SH.SPARQLRule
SH_TripleRule = SH.TripleRule
SH_SPARQLTarget = SH.SPARQLTarget
SH_SPARQLTargetType = SH.SPARQLTargetType
SH_JSTarget = SH.JSTarget
SH_JSTargetType = SH.JSTargetType
SH_JSFunction = SH.JSFunction

# predicates
RDF_type = RDF.type
RDF_first = RDF.first
RDF_rest = RDF.rest
RDF_object = RDF.object
RDF_predicate = RDF.predicate
RDF_subject = RDF.subject
RDFS_subClassOf = RDFS.subClassOf
RDFS_comment = RDFS.comment
SH_path = SH.path
SH_deactivated = SH.deactivated
SH_message = SH.message
SH_name = SH.name
SH_description = SH.description
SH_property = SH.property
SH_node = SH.node
SH_target = SH.target  # from advanced spec
SH_targetClass = SH.targetClass
SH_targetNode = SH.targetNode
SH_targetObjectsOf = SH.targetObjectsOf
SH_targetSubjectsOf = SH.targetSubjectsOf
SH_focusNode = SH.focusNode
SH_resultSeverity = SH.resultSeverity
SH_resultPath = SH.resultPath
SH_resultMessage = SH.resultMessage
SH_sourceConstraint = SH.sourceConstraint
SH_sourceConstraintComponent = SH.sourceConstraintComponent
SH_sourceShape = SH.sourceShape
SH_severity = SH.severity
SH_value = SH.value
SH_conforms = SH.conforms
SH_result = SH.result
SH_inversePath = SH.inversePath
SH_alternativePath = SH.alternativePath
SH_zeroOrMorePath = SH.zeroOrMorePath
SH_oneOrMorePath = SH.oneOrMorePath
SH_zeroOrOnePath = SH.zeroOrOnePath
SH_prefixes = SH.prefixes
SH_prefix = SH.prefix
SH_namespace = SH.namespace
SH_rule = SH.rule
SH_condition = SH.condition
SH_order = SH.order
SH_construct = SH.construct
SH_subject = SH.subject
SH_predicate = SH.predicate
SH_object = SH.object
SH_parameter = SH.parameter
SH_ask = SH.ask
SH_select = SH.select
SH_this = SH.this
SH_filterShape = SH.filterShape
SH_nodes = SH.nodes
SH_union = SH.union
SH_intersection = SH.intersection
SH_datatype = SH.datatype
SH_nodeKind = SH.nodeKind
SH_optional = SH.optional
SH_js = SH.js
SH_jsFunctionName = SH.jsFunctionName
SH_jsLibrary = SH.jsLibrary



import pprint

from decimal import Decimal
from typing import Union

import pyduktape2

from pyduktape2 import JSProxy
from rdflib import BNode, Literal, URIRef
from rdflib.namespace import XSD

from pyshacl.errors import ReportableRuntimeError

from . import load_into_context


class URIRefNativeWrapper(object):
    inner_type = "URIRef"

    def __init__(self, uri):
        if isinstance(uri, URIRef):
            self.inner = uri
        else:
            self.inner = URIRef(uri)

    @property
    def uri(self):
        return str(self.inner)

    def __eq__(self, other):
        return self.inner.__eq__(other.inner)

    def __repr__(self):
        inner_repr = repr(self.inner)
        return "URIRefNativeWrapper({})".format(inner_repr)


class BNodeNativeWrapper(object):
    inner_type = "BNode"

    def __init__(self, id_):
        if isinstance(id_, BNode):
            self.inner = id_
        else:
            self.inner = BNode(id_ or None)

    @property
    def identifier(self):
        return str(self.inner)

    def __eq__(self, other):
        return self.inner.__eq__(other.inner)

    def __repr__(self):
        inner_repr = repr(self.inner)
        return "BNodeNativeWrapper({})".format(inner_repr)


class LiteralNativeWrapper(object):
    inner_type = "Literal"

    def __init__(self, lexical, dtype=None, lang=None):
        if isinstance(lexical, Literal):
            self.inner = lexical
        else:
            if isinstance(dtype, URIRefNativeWrapper):
                dtype = dtype.inner
            self.inner = Literal(lexical, lang=lang, datatype=dtype)

    @property
    def lexical(self):
        return self.inner.value

    @property
    def language(self):
        return self.inner.language

    @property
    def datatype(self):
        return self.inner.datatype

    def __eq__(self, other):
        return self.inner.__eq__(other.inner)

    def __repr__(self):
        inner_repr = repr(self.inner)
        return "LiteralNativeWrapper({})".format(inner_repr)


class GraphNativeWrapper(object):
    def __init__(self, g):
        self.inner = g


class IteratorNativeWrapper(object):
    def __init__(self, it):
        self.it = it

    def it_next(self):
        return next(self.it)


def _make_uriref(args):
    uri = getattr(args, '0')
    return URIRefNativeWrapper(uri)


def _make_bnode(args):
    id_ = getattr(args, '0')
    return BNodeNativeWrapper(id_)


def _make_literal(args):
    lexical, dtype, lang = getattr(args, '0'), getattr(args, '1'), getattr(args, '2')
    if isinstance(dtype, JSProxy):
        as_native = getattr(dtype, '_native', None)
        if as_native is not None:
            dtype = as_native
    return LiteralNativeWrapper(lexical, dtype, lang)


def _native_node_equals(args):
    this, other = getattr(args, '0'), getattr(args, '1')
    if isinstance(this, (URIRefNativeWrapper, BNodeNativeWrapper, LiteralNativeWrapper)):
        this = this.inner
    if isinstance(other, (URIRefNativeWrapper, BNodeNativeWrapper, LiteralNativeWrapper)):
        other = other.inner
    return this == other


def _native_graph_find(args):
    # args are: g, s, p, o
    triple = [getattr(args, '0'), getattr(args, '1'), getattr(args, '2'), getattr(args, '3')]
    wrapped_triples = []
    for t in triple:
        if isinstance(t, (GraphNativeWrapper, URIRefNativeWrapper, BNodeNativeWrapper, LiteralNativeWrapper)):
            wrapped_triples.append(t.inner)
        else:
            wrapped_triples.append(t)
    g, s, p, o = wrapped_triples[:4]
    it = iter(g.triples((s, p, o)))
    return IteratorNativeWrapper(it)


def _native_iterator_next(args):
    arg0 = getattr(args, "0")
    if isinstance(arg0, IteratorNativeWrapper):
        arg0 = arg0.it
    try:
        spo_list = next(arg0)
    except StopIteration:
        return None
    wrapped_list = []
    for item in spo_list[:3]:
        if isinstance(item, URIRef):
            wrapped_list.append(URIRefNativeWrapper(item))
        elif isinstance(item, BNode):
            wrapped_list.append(BNodeNativeWrapper(item))
        elif isinstance(item, Literal):
            wrapped_list.append(LiteralNativeWrapper(item))
        else:
            raise RuntimeError("Bad item returned from iterator!")
    return wrapped_list


def _pprint(args):
    arg0 = getattr(args, '0')
    pprint.pprint(arg0)


printJs = '''\
function pprint(o) {
    return _pprint({'0': o});
}
'''

namedNodeJs = '''\
function NamedNode(uri, _native) {
    this.uri = uri;
    if (! _native) {
        _native = _make_uriref({'0': uri});
    }
    this._native = _native;
}
NamedNode.from_native = function(native) {
    var uri = native.uri;
    return new NamedNode(uri, native);
}
NamedNode.prototype.toPython = function() { return this._native; }
NamedNode.prototype.toString = function() { return "NamedNode("+this.uri+")"; }
NamedNode.prototype.isURI = function() { return true; }
NamedNode.prototype.isBlankNode = function() { return false; }
NamedNode.prototype.isLiteral = function() { return false; }
NamedNode.prototype.equals = function(other) {
    if (other.constructor && other.constructor === NamedNode) {
        return _native_node_equals({"0": this._native, "1": other._native});
    }
    return false;
}
'''

blankNodeJs = '''\
function BlankNode(id, _native) {
    this.id = id || null;
    if (! _native) {
        _native = _make_bnode({'0': id});
    }
    this._native = _native;
}
BlankNode.from_native = function(native) {
    var id = native.identifier;
    return new BlankNode(id, native);
}
BlankNode.prototype.toPython = function() { return this._native; }
BlankNode.prototype.toString = function() { return "BlankNode("+this.id+")"; }
BlankNode.prototype.isURI = function() { return false; }
BlankNode.prototype.isBlankNode = function() { return true; }
BlankNode.prototype.isLiteral = function() { return false; }
BlankNode.prototype.equals = function(other) {
    if (other.constructor && other.constructor === BlankNode) {
        return _native_node_equals({"0": this._native, "1": other._native});
    }
    return false;
}
'''

literalJs = '''\
function Literal(lex, languageOrDatatype, _native) {
    this.lex = lex;
    var ndict;
    if (languageOrDatatype.isURI && languageOrDatatype.isURI()) {
        this.language = "";  // Language cannot be null, be empty string
        this.datatype = languageOrDatatype;
        ndict = {'0': lex, '1': languageOrDatatype, '2': null};
    } else {
        var _lang = ""+languageOrDatatype;
        this.language = _lang;
        this.datatype = new NamedNode("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString");
        ndict = {'0': lex, '1': null, '2': languageOrDatatype};
    }
    if (! _native) {
        _native = _make_literal(ndict);
    }
    this._native = _native;
}
Literal.from_native = function(native) {
    var lex = native.lexical;
    var languageOrDatatype;
    var lang = native.language;
    var dt = native.datatype;
    if (lang) {
        languageOrDatatype = ""+lang;
    } else if (dt) {
        languageOrDatatype = new NamedNode(dt);
    } else {
        languageOrDatatype = new NamedNode("http://www.w3.org/2001/XMLSchema#string");
    }
    return new Literal(lex, languageOrDatatype, native);
}
Literal.prototype.toPython = function() { return this._native; }
Literal.prototype.toString = function() { return "Literal("+this.lex+", dt="+this.datatype+", lang='"+this.language+"')"; }
Literal.prototype.isURI = function() { return false; }
Literal.prototype.isBlankNode = function() { return false; }
Literal.prototype.isLiteral = function() { return true; }
Literal.prototype.equals = function(other) {
    if (other.constructor && other.constructor === Literal) {
        return _native_node_equals({'0': this._native, '1': other._native});
    }
    return false;
}
'''

termFactoryJs = '''\
function TermFactoryFactory() {
}
TermFactoryFactory.prototype.namedNode = function(uri) { return new NamedNode(uri); }
TermFactoryFactory.prototype.blankNode = function(id) { return new BlankNode(id); }
TermFactoryFactory.prototype.literal = function(lex, languageOrDatatype) { return new Literal(lex, languageOrDatatype); }
var TermFactory = new TermFactoryFactory();
'''

graphJs = '''\
function Triple(s, p, o) {
    this.subject = s;
    this.predicate = p;
    this.object = o;
}
function Iterator() {
    this._native = null;
    this.closed = false;
}
Iterator.from_native = function(native) {
    var i = new Iterator();
    i._native = native;
    return i;
}
Iterator.prototype.next = function() {
    if (this.closed) {
        1/0; //exception
    }
    if (this._native === null) {
        return null;
    }
    var bits = _native_iterator_next({"0": this._native});
    if (bits === null) {
        //This is the end of the iteration
        return null;
    }
    var converted_bits = [];
    for (var i=0; i<3; i++) {
        var b = bits[i];
        var inner_type = b.inner_type;
        if (inner_type === "URIRef") {
            converted_bits[i] = NamedNode.from_native(b);
        } else if (inner_type === "BNode") {
            converted_bits[i] = BlankNode.from_native(b);
        } else if (inner_type === "Literal") {
            converted_bits[i] = Literal.from_native(b);
        } else {
            1/0; //exception
        }
    }
    return new Triple(converted_bits[0], converted_bits[1], converted_bits[2]);
}
Iterator.prototype.close = function() {
    if (this.closed) { return; }
    this.closed = true;
    this._native = null;
}
Iterator.prototype.toPython = function() { return this._native; }
function Graph(_native) {
    this._native = _native;
}
Graph.prototype.toPython = function() { return this._native; }
Graph.prototype.find = function(s, p, o) {
    if (this._native === null) {
        return null;
    }
    if (s && s.hasOwnProperty('_native')) { s = s._native; }
    if (p && p.hasOwnProperty('_native')) { p = p._native; }
    if (o && o.hasOwnProperty('_native')) { o = o._native; }
    var native_it = _native_graph_find({'0': this._native, '1': s, '2': p, '3': o});
    var it = Iterator.from_native(native_it);
    return it;
}
'''


class SHACLJSContext(object):
    __slots__ = ("context", "fns")

    def __init__(self, data_graph, *args, shapes_graph=None, **kwargs):
        context = pyduktape2.DuktapeContext()
        context.set_globals(
            _pprint=_pprint,
            _make_uriref=_make_uriref,
            _make_bnode=_make_bnode,
            _make_literal=_make_literal,
            _native_node_equals=_native_node_equals,
            _native_iterator_next=_native_iterator_next,
            _native_graph_find=_native_graph_find,
        )
        context.eval_js(printJs)
        context.eval_js(termFactoryJs)
        context.eval_js(namedNodeJs)
        context.eval_js(blankNodeJs)
        context.eval_js(literalJs)
        context.eval_js(graphJs)

        context.set_globals(_native_data_graph=GraphNativeWrapper(data_graph))
        if shapes_graph is not None:
            context.set_globals(_native_shapes_graph=GraphNativeWrapper(shapes_graph))
        else:
            context.set_globals(_native_shapes_graph=None)
        context.eval_js('''var $data = new Graph(_native_data_graph);\n''')
        if shapes_graph is not None:
            context.eval_js('''var $shapes = new Graph(_native_shapes_graph);\n''')
        else:
            context.eval_js('''var $shapes;\n''')  # leave $shapes undefined
        context.set_globals(*args, **kwargs)

        self.context = context
        self.fns = {}

    def load_js_library(self, library: str):
        fns = load_into_context(self.context, library)
        self.fns.update(fns)

    @classmethod
    def build_results_as_constraint(cls, res):
        if isinstance(res, JSProxy):
            try:
                return res.toPython()
            except AttributeError:
                pass
            # this means its a JS Array or Object
            keys = list(iter(res))
            if len(keys) < 1:
                res = []
            else:
                first_key = keys[0]
                if isinstance(first_key, JSProxy):
                    # res is an array of objects
                    new_res = []
                    for k in keys:
                        try:
                            new_res.append(k.toPython())
                            continue
                        except AttributeError:
                            pass

                        v = getattr(k, 'value', None)
                        m = getattr(k, 'message', None)
                        p = getattr(k, 'path', None)

                        if v is not None:
                            try:
                                v = v.toPython()
                            except AttributeError:
                                pass
                        if v is not None and hasattr(v, 'inner'):
                            v = v.inner

                        if p is not None:
                            try:
                                p = p.toPython()
                            except AttributeError:
                                pass
                        if p is not None and hasattr(p, 'inner'):
                            p = p.inner
                        r = {'value': v, 'message': m, 'path': p}
                        new_res.append(r)
                    return new_res
                try:
                    getattr(res, first_key)
                    new_res = {}
                    for k in keys:
                        v = getattr(res, k, None)
                        if v is not None:
                            try:
                                v = v.toPython()
                            except AttributeError:
                                pass
                        if v is not None and hasattr(v, 'inner'):
                            v = v.inner
                        new_res[k] = v
                    return new_res
                except AttributeError:
                    # This must be an array of something else
                    res = keys
        elif isinstance(res, (Literal, URIRef, BNode)):
            pass
        elif isinstance(res, Decimal):
            res = Literal(res, datatype=XSD['decimal'])
        elif isinstance(res, (int, float, str, bytes, bool)):
            res = Literal(res)
        return res

    @classmethod
    def build_results_as_construct(cls, res):
        if isinstance(res, JSProxy):
            try:
                return res.toPython()
            except AttributeError:
                pass
            # this means its a JS Array or Object
            keys = list(iter(res))
            if len(keys) < 1:
                res = []
            else:
                first_key = keys[0]
                if isinstance(first_key, JSProxy):
                    # res is an array of objects or array of arrays
                    new_res = []
                    for k in keys:
                        try:
                            new_res.append(k.toPython())
                            continue
                        except AttributeError:
                            pass
                        subkeys = list(iter(k))
                        if len(subkeys) < 1:
                            new_res.append([])
                        else:
                            first_subkey = subkeys[0]
                            if isinstance(first_subkey, JSProxy):
                                if len(subkeys) >= 3:
                                    # res is an array of arrays
                                    this_s = subkeys[0]
                                    this_p = subkeys[1]
                                    this_o = subkeys[2]
                                else:
                                    raise ReportableRuntimeError(
                                        "JS Function returned incorrect number of items in the array."
                                    )
                            else:
                                this_s = getattr(k, 'subject', None)
                                this_p = getattr(k, 'predicate', None)
                                this_o = getattr(k, 'object', None)
                            try:
                                this_s = this_s.toPython()
                            except AttributeError:
                                pass
                            if this_s is not None and hasattr(this_s, 'inner'):
                                this_s = this_s.inner
                            try:
                                this_p = this_p.toPython()
                            except AttributeError:
                                pass
                            if this_p is not None and hasattr(this_p, 'inner'):
                                this_p = this_p.inner
                            try:
                                this_o = this_o.toPython()
                            except AttributeError:
                                pass
                            if this_o is not None and hasattr(this_o, 'inner'):
                                this_o = this_o.inner
                            new_res.append((this_s, this_p, this_o))
                    return new_res
                else:
                    # a JS Rules function must return an array of arrays, or array of objects
                    # otherwise, it does nothing!
                    return []
        return res

    @classmethod
    def build_results_as_target(cls, res):
        if isinstance(res, JSProxy):
            try:
                return res.toPython()
            except AttributeError:
                pass
            # this means its a JS Array or Object
            keys = list(iter(res))
            if len(keys) < 1:
                res = []
            else:
                first_key = keys[0]
                if isinstance(first_key, JSProxy):
                    # res is an array of objects or array of arrays
                    new_res = []
                    for k in keys:
                        try:
                            k = k.toPython()
                        except AttributeError:
                            pass
                        if k is not None and hasattr(k, 'inner'):
                            k = k.inner
                        new_res.append(k)
                    return new_res
                else:
                    # a JS Target function must return an array of Nodes
                    # otherwise, it does nothing!
                    return []
        return res

    @classmethod
    def build_results_as_shacl_function(cls, res, return_type=None):
        if isinstance(res, JSProxy):
            try:
                return res.toPython()
            except AttributeError:
                return None
        elif isinstance(res, (Literal, URIRef, BNode)):
            pass
        elif return_type is not None and isinstance(res, (int, float)):
            lex = str(res)
            res = Literal(lex, datatype=return_type, normalize=False)
        elif isinstance(res, (float, Decimal)):
            res = Literal(str(res), datatype=XSD['decimal'])
        elif isinstance(res, bool):
            res = Literal(res, datatype=XSD['boolean'])
        elif isinstance(res, (int, float, str, bytes, bool)):
            res = Literal(res, datatype=return_type, normalize=False)
        return res

    def get_fn_args(self, fn_name, args_map):
        c = self.context
        try:
            fn = c.get_global(fn_name)
        except BaseException as e:
            print(e)
            raise
        if fn is None:
            raise ReportableRuntimeError("JS Function {} cannot be found in the loaded files.".format(fn_name))
        if fn_name not in self.fns:
            raise ReportableRuntimeError("JS Function {} args cannot be determined. Bad JS structure?".format(fn_name))
        known_fn_args = self.fns[fn_name]  # type: tuple
        needed_args = []
        for k, v in args_map.items():
            look_for = "$" + str(k)
            positions = []
            start = 0
            while True:
                try:
                    pos = known_fn_args.index(look_for, start)
                    positions.append(pos)
                except ValueError:
                    break
                start = pos + 1
            if not positions:
                continue
            for p in positions:
                needed_args.append((p, k, v))
        for i, a in enumerate(known_fn_args):
            if a.startswith("$"):
                a = a[1:]
            if a not in args_map:
                needed_args.append((i, a, None))
        needed_args = [v for p, k, v in sorted(needed_args)]
        return needed_args

    def run_js_function(self, fn_name, args, returns: list = None):
        if returns is None:
            returns = []
        c = self.context
        args_string = ""
        bind_dict = {}
        preamble = ""
        for i, a in enumerate(args):
            arg_name = "fn_arg_" + str(i + 1)
            if isinstance(a, URIRef):
                wrapped_a: Union[URIRefNativeWrapper, BNodeNativeWrapper, LiteralNativeWrapper] = URIRefNativeWrapper(
                    a
                )
                native_name = "_{}_native".format(arg_name)
                preamble += "var {} = NamedNode.from_native({});\n".format(arg_name, native_name)
                bind_dict[native_name] = wrapped_a
            elif isinstance(a, BNode):
                wrapped_a = BNodeNativeWrapper(a)
                native_name = "_{}_native".format(arg_name)
                preamble += "var {} = BlankNode.from_native({});\n".format(arg_name, native_name)
                bind_dict[native_name] = wrapped_a
            elif isinstance(a, Literal):
                wrapped_a = LiteralNativeWrapper(a)
                native_name = "_{}_native".format(arg_name)
                preamble += "var {} = Literal.from_native({});\n".format(arg_name, native_name)
                bind_dict[native_name] = wrapped_a
            elif a is None:  # this is how we set an undefined variable
                preamble += "var {};\n".format(arg_name)
            else:
                bind_dict[arg_name] = a

            args_string = args_string + arg_name + ","
        c.set_globals(**bind_dict)
        args_string = args_string.rstrip(',')
        c.eval_js(preamble)
        res = c.eval_js("\n{}({});".format(fn_name, args_string))
        returns_dict = {}
        for r in returns:
            try:
                returns_dict[r] = c.get_global(r)
            except BaseException as e:
                print(e)
                returns_dict[r] = None
        returns_dict['_result'] = res
        return returns_dict



#
#
import typing

from urllib import request


try:
    import regex
except ImportError:
    import re

    regex = re

if typing.TYPE_CHECKING:
    from pyduktape2 import DuktapeContext

JS_FN_RE1 = regex.compile(rb'function\s+([^ \n]+)\s*\((.*)\)\s*\{', regex.MULTILINE | regex.IGNORECASE)
JS_FN_RE2 = regex.compile(
    rb'(?:let|const|var)\s+([^ \n]+)\s*=\s*function\s*\((.*)\)\s*\{', regex.MULTILINE | regex.IGNORECASE
)


def get_js_from_web(url: str):
    """

    :param url:
    :type url: str
    :return:
    """
    headers = {
        'Accept': 'application/javascript, text/javascript, application/ecmascript, text/ecmascript,' 'text/plain'
    }
    r = request.Request(url, headers=headers)
    resp = request.urlopen(r)
    code = resp.getcode()
    if not (200 <= code <= 210):
        raise RuntimeError("Cannot pull JS Library URL from the web: {}, code: {}".format(url, str(code)))
    return resp


def get_js_from_file(filepath: str):
    if filepath.startswith("file://"):
        filepath = filepath[7:]
    f = open(filepath, "rb")
    return f


def extract_functions(content):
    fns = {}
    matches1 = regex.findall(JS_FN_RE1, content)
    for m in matches1:
        name = m[0].decode('utf-8')
        params = tuple(p.strip().decode('utf-8') for p in m[1].split(b',') if p)
        fns[name] = params
    matches2 = regex.findall(JS_FN_RE2, content)
    for m in matches2:
        name = m[0].decode('utf-8')
        params = tuple(p.strip().decode('utf-8') for p in m[1].split(b',') if p)
        fns[name] = params
    return fns


def load_into_context(context: 'DuktapeContext', location: str):
    f = None
    try:
        if location.startswith("http:") or location.startswith("https:"):
            f = get_js_from_web(location)
        else:
            f = get_js_from_file(location)
        contents = f.read()
    finally:
        if f:
            f.close()
    fns = extract_functions(contents)
    context.eval_js(contents)
    return fns



from rdflib import Graph
from pyshacl import validate, extras
shapes_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:LanguageExampleShape
	a sh:NodeShape ;
	sh:targetClass ex:Country ;
	sh:js [
		a sh:JSConstraint ;
		sh:message "Values are literals with German language tag." ;
		sh:jsLibrary [ sh:jsLibraryURL "file://test/resources/js/germanLabel.js" ] ;
		sh:jsFunctionName "validateGermanLabel" ;
	] .
'''

data_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://example.com/ex#> .

ex:ValidCountry a ex:Country ;
	ex:germanLabel "Spanien"@de .

ex:InvalidCountry a ex:Country ;
	ex:germanLabel "Spain"@en .
'''

extras.dev_mode = True

def test_js_constraint():
    s1 = Graph().parse(data=shapes_graph, format="turtle")
    g1 = Graph().parse(data=data_graph, format="turtle")
    conforms, result_graph, result_text = validate(g1, shacl_graph=s1, advanced=True, debug=True, js=True)
    assert not conforms

if __name__ == "__main__":
    test_js_constraint()



# -*- coding: utf-8 -*-
import itertools

from typing import TYPE_CHECKING, Tuple, cast

import rdflib

from pyshacl.consts import SH_object, SH_predicate, SH_subject
from pyshacl.errors import ReportableRuntimeError
from pyshacl.helper.expression_helper import nodes_from_node_expression
from pyshacl.rules.shacl_rule import SHACLRule


if TYPE_CHECKING:
    from rdflib.term import Node

    from pyshacl.pytypes import GraphLike
    from pyshacl.shape import Shape


class TripleRule(SHACLRule):
    __slots__ = ("s", "p", "o")

    def __init__(self, shape: 'Shape', rule_node: 'rdflib.term.Identifier', **kwargs):
        """

        :param shape:
        :type shape: Shape
        :param rule_node:
        :type rule_node: rdflib.term.Identifier
        """
        super(TripleRule, self).__init__(shape, rule_node, **kwargs)
        my_subject_nodes = set(self.shape.sg.objects(self.node, SH_subject))
        if len(my_subject_nodes) < 1:
            raise RuntimeError("No sh:subject")
        elif len(my_subject_nodes) > 1:
            raise RuntimeError("Too many sh:subject")
        self.s = next(iter(my_subject_nodes))

        my_predicate_nodes = set(self.shape.sg.objects(self.node, SH_predicate))
        if len(my_predicate_nodes) < 1:
            raise RuntimeError("No sh:predicate")
        elif len(my_predicate_nodes) > 1:
            raise RuntimeError("Too many sh:predicate")
        self.p = next(iter(my_predicate_nodes))

        my_object_nodes = set(self.shape.sg.objects(self.node, SH_object))
        if len(my_object_nodes) < 1:
            raise RuntimeError("No sh:object")
        elif len(my_object_nodes) > 1:
            raise RuntimeError("Too many sh:object")
        self.o = next(iter(my_object_nodes))

    def apply(self, data_graph: 'GraphLike') -> int:
        focus_nodes = self.shape.focus_nodes(data_graph)  # uses target nodes to find focus nodes
        applicable_nodes = self.filter_conditions(focus_nodes, data_graph)
        all_added = 0
        iterate_limit = 100
        while True:
            if iterate_limit < 1:
                raise ReportableRuntimeError("sh:rule iteration exceeded iteration limit of 100.")
            iterate_limit -= 1
            added = 0
            to_add = []
            for a in applicable_nodes:
                s_set = nodes_from_node_expression(self.s, a, data_graph, self.shape.sg)
                p_set = nodes_from_node_expression(self.p, a, data_graph, self.shape.sg)
                o_set = nodes_from_node_expression(self.o, a, data_graph, self.shape.sg)
                new_triples = itertools.product(s_set, p_set, o_set)
                this_added = False
                for i in iter(new_triples):
                    if not this_added and i not in data_graph:
                        this_added = True
                    to_add.append(i)
                if this_added:
                    added += 1
            if added > 0:
                for i in to_add:
                    data_graph.add(cast(Tuple['Node', 'Node', 'Node'], i))
                all_added += added
                if self.iterate:
                    continue  # Jump up to iterate
                else:
                    break  # Don't iterate
            break
        return all_added



from rdflib import Graph
from pyshacl import validate
shapes_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://datashapes.org/js/tests/rules/rectangle.test#> .

ex:RectangleShape
	a sh:NodeShape ;
	sh:targetClass ex:Rectangle ;
	sh:rule [
		a sh:JSRule ;    # This triple is optional
		sh:jsFunctionName "computeArea" ;
		sh:jsLibrary [ sh:jsLibraryURL "file://test/resources/js/rectangle.js"^^xsd:anyURI ] ;
    ] ;
    sh:property [
        sh:path ex:area ;
        sh:datatype xsd:double ;
        sh:minCount 1 ;
        sh:maxCount 1 ;
    ] .
'''

data_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix ex: <http://datashapes.org/js/tests/rules/rectangle.test#> .
@prefix exdata: <http://datashapes.org/js/tests/rules/rectangle.test.data#> .

exdata:ExampleRectangle
	a ex:Rectangle ;
	ex:width 7 ;
	ex:height 8 .
'''

def test_js_rules():
    s1 = Graph().parse(data=shapes_graph, format="turtle")
    g1 = Graph().parse(data=data_graph, format="turtle")
    conforms, result_graph, result_text = validate(g1, shacl_graph=s1, advanced=True, debug=True, js=True)
    assert not conforms

if __name__ == "__main__":
    test_js_rules()



# -*- coding: utf-8 -*-
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, Union

from pyshacl.consts import RDF_type, SH_rule, SH_SPARQLRule, SH_TripleRule
from pyshacl.errors import ReportableRuntimeError, RuleLoadError
from pyshacl.pytypes import GraphLike
from pyshacl.rules.sparql import SPARQLRule
from pyshacl.rules.triple import TripleRule


if TYPE_CHECKING:
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph

    from .shacl_rule import SHACLRule


def gather_rules(shacl_graph: 'ShapesGraph', iterate_rules=False) -> Dict['Shape', List['SHACLRule']]:
    """

    :param shacl_graph:
    :type shacl_graph: ShapesGraph
    :return:
    :rtype: Dict[Shape, List[SHACLRule]]
    """
    triple_rule_nodes = set(shacl_graph.subjects(RDF_type, SH_TripleRule))
    sparql_rule_nodes = set(shacl_graph.subjects(RDF_type, SH_SPARQLRule))
    if shacl_graph.js_enabled:
        from pyshacl.extras.js.rules import JSRule, SH_JSRule

        js_rule_nodes = set(shacl_graph.subjects(RDF_type, SH_JSRule))
        use_JSRule: Union[bool, Type] = JSRule
    else:
        use_JSRule = False
        js_rule_nodes = set()
    overlaps = triple_rule_nodes.intersection(sparql_rule_nodes)
    if len(overlaps) > 0:
        raise RuleLoadError(
            "A SHACL Rule cannot be both a TripleRule and a SPARQLRule.",
            "https://www.w3.org/TR/shacl-af/#rules-syntax",
        )
    overlaps = triple_rule_nodes.intersection(js_rule_nodes)
    if len(overlaps) > 0:
        raise RuleLoadError(
            "A SHACL Rule cannot be both a TripleRule and a JSRule.",
            "https://www.w3.org/TR/shacl-af/#rules-syntax",
        )
    overlaps = sparql_rule_nodes.intersection(js_rule_nodes)
    if len(overlaps) > 0:
        raise RuleLoadError(
            "A SHACL Rule cannot be both a SPARQLRule and a JSRule.",
            "https://www.w3.org/TR/shacl-af/#rules-syntax",
        )
    used_rules = shacl_graph.subject_objects(SH_rule)
    ret_rules = defaultdict(list)
    for sub, obj in used_rules:
        try:
            shape: Shape = shacl_graph.lookup_shape_from_node(sub)
        except (AttributeError, KeyError):
            raise RuleLoadError(
                "The shape that rule is attached to is not a valid SHACL Shape.",
                "https://www.w3.org/TR/shacl-af/#rules-syntax",
            )
        if obj in triple_rule_nodes:
            rule: SHACLRule = TripleRule(shape, obj, iterate=iterate_rules)
        elif obj in sparql_rule_nodes:
            rule = SPARQLRule(shape, obj)
        elif use_JSRule and callable(use_JSRule) and obj in js_rule_nodes:
            rule = use_JSRule(shape, obj)
        else:
            raise RuleLoadError(
                "when using sh:rule, the Rule must be defined as either a TripleRule or SPARQLRule.",
                "https://www.w3.org/TR/shacl-af/#rules-syntax",
            )
        ret_rules[shape].append(rule)
    return ret_rules


def apply_rules(shapes_rules: Dict, data_graph: GraphLike, iterate=False) -> int:
    # short the shapes dict by shapes sh:order before execution
    sorted_shapes_rules: List[Tuple[Any, Any]] = sorted(shapes_rules.items(), key=lambda x: x[0].order)
    total_modified = 0
    for shape, rules in sorted_shapes_rules:
        # sort the rules by the sh:order before execution
        rules = sorted(rules, key=lambda x: x.order)
        iterate_limit = 100
        while True:
            if iterate_limit < 1:
                raise ReportableRuntimeError("SHACL Shape Rule iteration exceeded iteration limit of 100.")
            iterate_limit -= 1
            this_modified = 0
            for r in rules:
                if r.deactivated:
                    continue
                n_modified = r.apply(data_graph)
                this_modified += n_modified
            if this_modified > 0:
                total_modified += this_modified
                if iterate:
                    continue
                else:
                    break
            break
    return total_modified



# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/133
"""

import rdflib

from pyshacl import validate


shacl_file = r"""@prefix geo: <http://www.opengis.net/ont/geosparql#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@base <http://www.opengis.net/def/geosparql/validator/> .

<S16-wkt-content>
	a sh:NodeShape ;
	sh:property <S16-wkt-content-sub-start> ;
	sh:targetSubjectsOf geo:asWKT ;
.

<S16-wkt-content-sub-start>
	a sh:PropertyShape ;
	sh:path geo:asWKT ;
	sh:pattern "^\\s*$|^\\s*(P|C|S|L|T|<)" ;
	sh:flags "i" ;
	sh:message "The content of an RDF literal with an incoming geo:asWKT relation must conform to a well-formed WKT string, as defined by its official specification (Simple Features Access)."@en ;
.
"""


data_file = """@prefix geo: <http://www.opengis.net/ont/geosparql#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .

<https://example.com/geometry-a>
    geo:asWKT "POINT (153.084231 -27.322738)"^^geo:wktLiteral ;
.

<https://example.com/geometry-b>
    geo:asWKT "xPOINT (153.084231 -27.322738)"^^geo:wktLiteral ;
.

<https://example.com/geometry-c>
    geo:asWKT "(153.084231 -27.322738)"^^geo:wktLiteral ;
.

<https://example.com/geometry-d>
    geo:asWKT "     POINT (153.084231 -27.322738)"^^geo:wktLiteral ;
.

<https://example.com/geometry-e>
    geo:asWKT "     "^^geo:wktLiteral ;
.

<https://example.com/geometry-f>
    geo:asWKT ""^^geo:wktLiteral ;
.
"""


def test_133() -> None:
    data = rdflib.Graph()
    data.parse(data=data_file, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    shapes.print()
    res = validate(
        data,
        shacl_graph=shapes,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        debug=True,
    )
    conforms, graph, string = res
    assert False == conforms
    assert "geometry-b" in string
    assert "geometry-c" in string
    assert "geometry-a" not in string
    assert "geometry-d" not in string
    assert "geometry-e" not in string
    assert "geometry-f" not in string


if __name__ == "__main__":
    test_133()



# -*- coding: utf-8 -*-
#
import os
import pickle
import platform
import sys

from io import BufferedIOBase, BytesIO, TextIOBase, UnsupportedOperation
from pathlib import Path
from typing import IO, BinaryIO, List, Optional, Union, cast
from urllib import request
from urllib.error import HTTPError

import rdflib

from .clone import clone_dataset, clone_graph


ConjunctiveLike = Union[rdflib.ConjunctiveGraph, rdflib.Dataset]
GraphLike = Union[ConjunctiveLike, rdflib.Graph]


is_windows = platform.system() == "Windows"


baked_in = {}


def add_baked_in(url, graph_path):
    baked_in[url] = graph_path


def get_rdf_from_web(url: Union[rdflib.URIRef, str]):
    """

    :param url:
    :type url: rdflib.URIRef | str
    :return:
    """
    nohash = url.rstrip("#")
    if nohash in baked_in:
        g = baked_in[nohash]
        if g[-7:] == ".pickle":
            with open(g, 'rb') as g_pickle:
                u = pickle.Unpickler(g_pickle, fix_imports=False)
                g_store, identifier = u.load()
            graph = rdflib.Graph(store=g_store, identifier=identifier)
            return graph, "graph"
        else:
            return g, None

    # Ask for everything we know about
    headers = {'Accept': 'text/turtle, application/rdf+xml, application/ld+json, application/n-triples, text/plain'}
    known_format = None

    r = request.Request(url, headers=headers)
    resp = request.urlopen(r)
    code = resp.getcode()
    if not (200 <= code <= 210):
        raise RuntimeError("Cannot pull RDF URL from the web: {}, code: {}".format(url, str(code)))

    content_type = resp.headers.get('Content-Type', None)
    if content_type:
        if content_type.startswith("text/turtle"):
            known_format = "turtle"
        elif content_type.startswith("application/rdf+xml"):
            known_format = "xml"
        elif content_type.startswith("application/xml"):
            known_format = "xml"
        elif content_type.startswith("application/ld+json"):
            known_format = "json-ld"
        elif content_type.startswith("application/n-triples"):
            known_format = "nt"
    return resp, known_format


def load_from_source(
    source: Union[GraphLike, BufferedIOBase, TextIOBase, BinaryIO, Union[str, bytes]],
    g: Optional[GraphLike] = None,
    rdf_format: Optional[str] = None,
    multigraph: bool = False,
    do_owl_imports: Union[bool, int] = False,
    import_chain: Optional[List[Union[rdflib.URIRef, str]]] = None,
):
    """

    :param source:
    :param g:
    :type g: rdflib.Graph | None
    :param rdf_format:
    :type rdf_format: str
    :param multigraph:
    :type multigraph: bool
    :param do_owl_imports:
    :type do_owl_imports: bool|int
    :param import_chain:
    :type import_chain: list | None
    :return:
    """
    source_is_graph = False
    open_source: Optional[Union[BufferedIOBase, BinaryIO]] = None
    source_was_open: bool = False
    source_as_file: Optional[Union[BufferedIOBase, BinaryIO]] = None
    source_as_filename: Optional[str] = None
    source_as_bytes: Optional[bytes] = None
    filename = None
    public_id = None
    uri_prefix = None
    is_imported_graph = do_owl_imports and isinstance(do_owl_imports, int) and do_owl_imports > 1
    if isinstance(source, (rdflib.Graph, rdflib.ConjunctiveGraph, rdflib.Dataset)):
        source_is_graph = True
        if g is None:
            g = source
        else:
            raise RuntimeError("Cannot pass in both target=rdflib.Graph/Dataset and g=graph.")
    elif isinstance(source, (BufferedIOBase, TextIOBase)):
        if hasattr(source, 'name'):
            filename = source.name  # type: ignore
            public_id = Path(filename).resolve().as_uri() + "#"
        if isinstance(source, TextIOBase):
            buf = getattr(source, "buffer")  # type: BufferedIOBase
            source_as_file = source = buf
        else:
            source_as_file = source
        if hasattr(source, 'closed'):
            if not bool(source.closed):
                open_source = source
                source_was_open = True
        else:
            # Assume it is open now and it was open when we started.
            open_source = source
            source_was_open = True

    elif isinstance(source, str):
        if source == "stdin" or source == "-" or source == "/dev/stdin":
            public_id = "/dev/stdin"
            filename = "/dev/stdin"
            source_as_filename = filename
        elif is_windows and source.startswith('file:///'):
            public_id = source
            filename = source[8:]
            source_as_filename = filename
        elif not is_windows and source.startswith('file://'):
            public_id = source
            filename = source[7:]
            source_as_filename = filename
        elif source.startswith('http:') or source.startswith('https:'):
            public_id = source
            try:
                resp, rdf_format = get_rdf_from_web(source)
            except HTTPError:
                if is_imported_graph:
                    return g
                else:
                    raise
            if rdf_format == 'graph':
                source = resp
                source_is_graph = True
            else:
                filename = resp.geturl()
                fp = resp.fp  # type: BufferedIOBase
                source_was_open = False
                source = open_source = fp
        else:
            first_char = source[0]
            if is_windows and (first_char == '\\' or (len(source) > 3 and source[1:3] == ":\\")):
                filename = source
                source_as_filename = filename
            elif first_char == '/' or (len(source) > 2 and source[0:2] == "./"):
                filename = source
                source_as_filename = filename
            elif (
                first_char == '#'
                or first_char == '@'
                or first_char == '<'
                or first_char == '\n'
                or first_char == '{'
                or first_char == '['
            ):
                # Contains some JSON or XML or Turtle chars, it's not a path
                source_as_file = None
                source_as_filename = None
            elif len(source) >= 32 and '\n' in source[:32]:
                # Contains a new line near the start of the file, can't be a path
                source_as_file = None
                source_as_filename = None
            elif len(source) < 140:
                filename = source
                source_as_filename = filename
        if source_as_filename and filename:
            pid = os.getpid()
            fd0 = "/proc/{}/fd/0".format(str(pid))
            if filename == "/dev/stdin" or filename == fd0:
                source = source_as_file = open_source = sys.stdin.buffer
                source_was_open = True
            else:
                try:
                    filename = os.readlink(filename)
                    if filename == fd0 or filename == "/dev/stdin":
                        source = source_as_file = open_source = sys.stdin.buffer
                        source_was_open = True
                except OSError:
                    pass
        # TODO: Do we still need this? Not sure why this was added, but works better without it
        #  if public_id and not public_id.endswith('#'):
        #     public_id = "{}#".format(public_id)
        if not source_as_file and not source_as_filename and not open_source and isinstance(source, str):
            # source is raw RDF data.
            source_as_bytes = source = source.encode('utf-8')
    elif isinstance(source, bytes):
        if source.startswith(b'file:') or source.startswith(b'http:') or source.startswith(b'https:'):
            raise ValueError("file:// and http:// strings should be given as str, not bytes.")
        first_char_b: bytes = source[0:1]
        if (
            first_char_b == b'#'
            or first_char_b == b'@'
            or first_char_b == b'<'
            or first_char_b == b'\n'
            or first_char_b == b'{'
            or first_char_b == b'['
        ):
            # Contains some JSON or XML or Turtle stuff
            source_as_file = None
            source_as_filename = None
        elif len(source) < 140:
            filename = source.decode('utf-8')
            source_as_filename = filename
        if not source_as_file and not source_as_filename and not open_source:
            source_as_bytes = source
    else:
        raise ValueError("Cannot determine the format of the input graph")
    if g is None:
        if source_is_graph:
            target_g: Union[rdflib.Graph, rdflib.ConjunctiveGraph, rdflib.Dataset] = source  # type: ignore
        else:
            target_g = rdflib.Dataset() if multigraph else rdflib.Graph()
    else:
        if not isinstance(g, (rdflib.Graph, rdflib.Dataset, rdflib.ConjunctiveGraph)):
            raise RuntimeError("Passing in 'g' must be a rdflib Graph or Dataset.")
        target_g = g

    if filename:
        if filename.endswith('.ttl'):
            rdf_format = rdf_format or 'turtle'
        elif filename.endswith('.nt'):
            rdf_format = rdf_format or 'nt'
        elif filename.endswith('.n3'):
            rdf_format = rdf_format or 'n3'
        elif filename.endswith('.json'):
            rdf_format = rdf_format or 'json-ld'
        elif filename.endswith('.nq') or filename.endswith('.nquads'):
            rdf_format = rdf_format or 'nquads'
        elif filename.endswith('.trig'):
            rdf_format = rdf_format or 'trig'
        elif filename.endswith('.xml') or filename.endswith('.rdf'):
            rdf_format = rdf_format or 'xml'
    if source_as_filename and filename is not None and not open_source:
        filename = str(Path(filename).resolve())
        if not public_id:
            public_id = Path(filename).as_uri() + "#"
        source = open_source = open(filename, mode='rb')
    if not open_source and source_as_bytes:
        source = open_source = BytesIO(source_as_bytes)  # type: ignore
    if open_source:
        _source = open_source
        # Check if we can seek
        try:
            _source.seek(0)  # type: ignore
        except (AttributeError, UnsupportedOperation):
            # Read it all into memory
            new_bytes = BytesIO(_source.read())
            if not source_was_open:
                _source.close()
            source = _source = new_bytes
            source_was_open = False
        if rdf_format is None:
            line = _source.readline().lstrip()
            line_len = len(line) if line is not None else 0
            while (
                (line is not None and line_len == 0)
                or (line_len == 1 and line[0] == "\n")
                or (line_len == 2 and line[0:2] == "\r\n")
            ):
                line = _source.readline().lstrip()
                line_len = len(line) if line is not None else 0
            if line_len > 15:
                line = line[:15]
            line = line.lower()
            if line.startswith(b"<!doctype html") or line.startswith(b"<html"):
                raise RuntimeError("Attempted to load a HTML document as RDF.")
            if line.startswith(b"<?xml") or line.startswith(b"<xml") or line.startswith(b"<rdf:"):
                rdf_format = "xml"
            if line.startswith(b"@base ") or line.startswith(b"@prefix ") or line.startswith(b"PREFIX "):
                rdf_format = "turtle"
            try:
                _source.seek(0)
            except (AttributeError, UnsupportedOperation):
                raise RuntimeError("Seek failed while identifying file type.")
            except ValueError:
                raise RuntimeError("File closed while identifying file type.")
        if rdf_format == 'turtle' or rdf_format == 'n3':
            # SHACL Shapes files and Data files can have extra RDF Metadata in the
            # Top header block, including #BaseURI and #Prefix.
            # The @base line is not read here, but it is parsed in the n3 parser
            while True:
                try:
                    line = _source.readline()
                    assert line is not None and len(line) > 0
                except AssertionError:
                    break
                # Strip line from start
                while len(line) > 0 and line[0:1] in b' \t\n\r\x0B\x0C\x85\xA0':
                    line = line[1:]
                # We reached the end of the line, check the next line
                if len(line) < 1:
                    continue
                # If this is not a comment, then this is the first non-comment line, we're done.
                if not line[0:1] == b'#':
                    break
                # Strip from start again, but now removing hashes too.
                while len(line) > 0 and line[0:1] in b'# \t\xA0':
                    line = line[1:]
                # Strip line from end
                while len(line) > 0 and line[-1:] in b' \t\n\r\x0B\x0C\x85\xA0':
                    line = line[:-1]
                spl = line.split(b':', 1)
                if len(spl) < 2:
                    continue
                keyword = spl[0].lower()
                # Strip keyword end
                while len(keyword) > 0 and keyword[-1:] in b' \t\n\r\x0B\x0C\x85\xA0':
                    keyword = keyword[:-1]
                if len(keyword) < 1:
                    continue
                wordval = spl[1]
                # Strip wordval start
                while len(wordval) > 0 and wordval[0:1] in b' \t\n\r\x0B\x0C\x85\xA0':
                    wordval = wordval[1:]
                if len(wordval) < 1:
                    continue
                wordval_str = wordval.decode('utf-8')
                if keyword == b"baseuri":
                    public_id = wordval_str
                elif keyword == b"prefix":
                    uri_prefix = wordval_str
            try:
                _source.seek(0)
            except (AttributeError, UnsupportedOperation):
                raise RuntimeError("Seek failed while pre-parsing Turtle File.")
            except ValueError:
                raise RuntimeError("File closed while pre-parsing Turtle File.")
        target_g.parse(source=cast(IO[bytes], _source), format=rdf_format, publicID=public_id)
        # If the target was open to begin with, leave it open.
        if not source_was_open:
            _source.close()
        elif hasattr(_source, 'seek'):
            try:
                _source.seek(0)
            except (AttributeError, UnsupportedOperation):
                pass
            except ValueError:
                # The parser closed our file!
                pass
        source_is_graph = True
    elif source_is_graph and (target_g != source):
        # clone source into g
        if isinstance(target_g, (rdflib.Dataset, rdflib.ConjunctiveGraph)) and isinstance(
            source, (rdflib.Dataset, rdflib.ConjunctiveGraph)
        ):
            clone_dataset(source, target_g)
        elif isinstance(target_g, rdflib.Graph) and isinstance(source, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            raise RuntimeError("Cannot load a Dataset source into a Graph target.")
        elif isinstance(target_g, (rdflib.Dataset, rdflib.ConjunctiveGraph)) and isinstance(source, rdflib.Graph):
            target = rdflib.Graph(store=target_g.store, identifier=public_id)
            clone_graph(source, target)
        elif isinstance(target_g, rdflib.Graph) and isinstance(source, rdflib.Graph):
            clone_graph(source, target_g)
        else:
            raise RuntimeError("Cannot merge source graph into target graph.")

    if not source_is_graph:
        raise RuntimeError("Error opening graph from source.")

    if public_id:
        if uri_prefix:
            if is_imported_graph and uri_prefix == '':
                # Don't reassign blank prefix, when importing subgraph
                pass
            else:
                has_named_prefix = target_g.store.namespace(uri_prefix)
                if not has_named_prefix:
                    target_g.namespace_manager.bind(uri_prefix, public_id)
        elif not is_imported_graph:
            existing_blank_prefix = target_g.store.namespace('')
            if not existing_blank_prefix:
                target_g.namespace_manager.bind('', public_id)
    if do_owl_imports:
        if isinstance(do_owl_imports, int):
            if do_owl_imports > 3:
                return target_g
        else:
            do_owl_imports = 1

        if import_chain is None:
            import_chain = []
        if public_id and (public_id.endswith('#') or public_id.endswith('/')):
            root_id: Union[rdflib.URIRef, None] = rdflib.URIRef(public_id[:-1])
        else:
            root_id = rdflib.URIRef(public_id) if public_id else None
        done_imports = 0
        if root_id is not None:
            if isinstance(target_g, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
                gs = list(target_g.contexts())
            else:
                gs = [target_g]
            for ng in gs:
                owl_imports = list(ng.objects(root_id, rdflib.OWL.imports))
                if len(owl_imports) > 0:
                    import_chain.append(str(root_id))
                for i in owl_imports:
                    imp_str = str(i)
                    if imp_str in import_chain:
                        continue
                    load_from_source(
                        imp_str,
                        g=target_g,
                        multigraph=multigraph,
                        do_owl_imports=do_owl_imports + 1,
                        import_chain=import_chain,
                    )
                    done_imports += 1
        if done_imports < 1 and public_id is not None and root_id != public_id:
            public_id_uri = rdflib.URIRef(public_id)
            if isinstance(target_g, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
                gs = list(target_g.contexts())
            else:
                gs = [target_g]
            for ng in gs:
                owl_imports = list(ng.objects(public_id_uri, rdflib.OWL.imports))
                if len(owl_imports) > 0:
                    import_chain.append(str(public_id_uri))
                for i in owl_imports:
                    imp_str = str(i)
                    if imp_str in import_chain:
                        continue
                    load_from_source(
                        imp_str,
                        g=target_g,
                        multigraph=multigraph,
                        do_owl_imports=do_owl_imports + 1,
                        import_chain=import_chain,
                    )
                    done_imports += 1
        if done_imports < 1:
            if isinstance(target_g, (rdflib.ConjunctiveGraph, rdflib.Dataset)):
                gs = list(target_g.contexts())
            else:
                gs = [target_g]
            for ng in gs:
                ontologies = ng.subjects(rdflib.RDF.type, rdflib.OWL.Ontology)
                for ont in ontologies:
                    if ont == root_id or ont == public_id:
                        continue
                    ont_str = str(ont)
                    if ont_str in import_chain:
                        continue
                    import_chain.append(ont_str)
                    owl_imports = list(ng.objects(ont, rdflib.OWL.imports))
                    for i in owl_imports:
                        imp_str = str(i)
                        if imp_str in import_chain:
                            continue
                        load_from_source(
                            imp_str,
                            g=target_g,
                            multigraph=multigraph,
                            do_owl_imports=do_owl_imports + 1,
                            import_chain=import_chain,
                        )
                        done_imports += 1
    return target_g



# -*- coding: utf-8 -*-
#
from typing import Optional, Union

import rdflib

from rdflib.collection import Collection

from .consts import RDF_first
from .pytypes import ConjunctiveLike, GraphLike


def clone_dataset(source_ds, target_ds=None):
    if target_ds and not isinstance(target_ds, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
        raise RuntimeError("when cloning a dataset, the target_ds must be a conjunctiveGraph or rdflib Dataset.")
    default_union = source_ds.default_union
    if target_ds is None:
        target_ds = rdflib.Dataset(default_union=default_union)
    named_graphs = [
        rdflib.Graph(source_ds.store, i, namespace_manager=source_ds.namespace_manager)
        if not isinstance(i, rdflib.Graph)
        else i
        for i in source_ds.store.contexts(None)
    ]
    cloned_graphs = [
        clone_graph(ng, rdflib.Graph(target_ds.store, ng.identifier, namespace_manager=target_ds.namespace_manager))
        for ng in named_graphs
    ]
    default_context_id = target_ds.default_context.identifier
    for g in cloned_graphs:
        if g.identifier == default_context_id:
            target_ds.store.remove_graph(target_ds.default_context)
            target_ds.default_context = g
        target_ds.add_graph(g)
    return target_ds


def clone_graph(source_graph, target_graph=None, identifier=None):
    """
    Make a clone of the source_graph by directly copying triples from source_graph to target_graph
    :param source_graph:
    :type source_graph: rdflib.Graph
    :param target_graph:
    :type target_graph: rdflib.Graph|None
    :param identifier:
    :type identifier: str | None
    :return: The cloned graph
    :rtype: rdflib.Graph
    """
    if isinstance(source_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
        return clone_dataset(source_graph, target_ds=target_graph)
    if target_graph is None:
        g = rdflib.Graph(identifier=identifier)
        for p, n in source_graph.namespace_manager.namespaces():
            g.namespace_manager.bind(p, n, override=True, replace=True)
    else:
        g = target_graph
        for p, n in source_graph.namespace_manager.namespaces():
            g.namespace_manager.bind(p, n, override=False, replace=False)
    for t in iter(source_graph):
        g.add(t)
    return g


def mix_datasets(
    base_ds: ConjunctiveLike, extra_ds: GraphLike, target_ds: Optional[Union[ConjunctiveLike, str]] = None
):
    """
    Make a clone of base_ds (dataset) and add in the triples from extra_ds (dataset)
    :param base_ds:
    :type base_ds: rdflib.Dataset
    :param extra_ds:
    :type extra_ds: rdflib.Dataset
    :param target_ds:
    :type target_ds: rdflib.Dataset|str|NoneType
    :return: The cloned Dataset with mixed in triples from extra_ds
    :rtype: rdflib.Dataset
    """
    default_union = base_ds.default_union
    base_named_graphs = list(base_ds.contexts())
    if target_ds is None:
        target_ds = rdflib.Dataset(default_union=default_union)
    elif isinstance(target_ds, rdflib.ConjunctiveGraph):
        raise RuntimeError("Cannot mix new graphs into a ConjunctiveGraph, use Dataset instead.")
    elif target_ds == "inplace":
        pass  # do nothing here
    elif not isinstance(target_ds, rdflib.Dataset):
        raise RuntimeError("Cannot mix datasets if target_ds passed in is not a Dataset itself.")
    if isinstance(extra_ds, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
        mixin_graphs = list(extra_ds.contexts())
    else:
        mixin_graphs = [extra_ds]
    if target_ds == "inplace":
        target_ds = base_ds
        for mg in mixin_graphs:
            mod_named_graphs = {g.identifier: mix_graphs(g, mg, target_graph="inplace") for g in base_named_graphs}
    elif isinstance(target_ds, str):
        raise RuntimeError("target_ds cannot be a string (unless it is 'inplace')")
    else:

        mixed_graphs = {}
        for mg in mixin_graphs:
            mod_named_graphs = {
                g.identifier: mix_graphs(
                    g, mg, target_graph=rdflib.Graph(store=target_ds.store, identifier=g.identifier)
                )
                for g in base_named_graphs
            }
            mixed_graphs.update(mod_named_graphs)
        default_context_id = target_ds.default_context.identifier
        for i, m in mixed_graphs.items():
            if i == default_context_id:
                target_ds.store.remove_graph(target_ds.default_context)
                target_ds.default_context = m
            target_ds.add_graph(m)
    return target_ds


def mix_graphs(base_graph: GraphLike, extra_graph: GraphLike, target_graph: Optional[Union[GraphLike, str]] = None):
    """
    Make a clone of base_graph and add in the triples from extra_graph
    :param base_graph:
    :type base_graph: rdflib.Graph
    :param extra_graph:
    :type extra_graph: rdflib.Graph
    :param target_graph:
    :type target_graph: rdflib.Graph|str|NoneType
    :return: The cloned graph with mixed in triples from extra_graph
    :rtype: rdflib.Graph
    """
    if isinstance(base_graph, (rdflib.ConjunctiveGraph, rdflib.Dataset)) and isinstance(
        target_graph, (rdflib.ConjunctiveGraph, rdflib.Dataset)
    ):
        return mix_datasets(base_graph, extra_graph, target_ds=target_graph)
    if target_graph is None:
        g = clone_graph(base_graph, target_graph=None, identifier=base_graph.identifier)
    elif target_graph == "inplace":
        # Special case, don't clone the basegraph, just put extra straight in
        g = base_graph
    elif isinstance(target_graph, str):
        raise RuntimeError("target_graph cannot be a string (unless it is 'inplace')")
    else:
        g = clone_graph(base_graph, target_graph=target_graph)
    g = clone_graph(extra_graph, target_graph=g)
    return g


def clone_list(graph, lnode, target_graph, keepid=False, recursion=0):
    if keepid:
        cloned_node = rdflib.BNode(str(lnode))
    else:
        cloned_node = rdflib.BNode()
    new_list = Collection(target_graph, cloned_node)
    for item in iter(graph.items(lnode)):
        cloned_item = clone_node(graph, item, target_graph, recursion=recursion + 1)
        new_list.append(cloned_item)
    return cloned_node


def clone_blank_node(graph, bnode, target_graph, keepid=False, recursion=0):
    if not isinstance(graph, rdflib.Graph):
        raise RuntimeError("clone_blank_node must take an rdflib.Graph as first parameter")
    if not isinstance(bnode, rdflib.BNode):
        raise RuntimeError("clone_blank_node must take an rdflib.BNode as second parameter")
    if keepid:
        cloned_bnode = rdflib.BNode(str(bnode))
    else:
        cloned_bnode = rdflib.BNode()
    if recursion >= 10:
        return cloned_bnode  # Cannot clone this deep

    predicates = set(graph.predicates(bnode))
    if len(predicates) < 1:
        return cloned_bnode
    if RDF_first in predicates:
        # don't increase recursion here, we're not actually going any deeper in the graph, just sideways
        return clone_list(graph, bnode, target_graph, keepid=keepid, recursion=recursion)
    for p in predicates:
        cloned_p = clone_node(graph, p, target_graph, recursion=recursion + 1)
        objs = list(graph.objects(bnode, p))
        if len(objs) < 1:
            continue
        for o in objs:
            cloned_o = clone_node(graph, o, target_graph, recursion=recursion + 1)
            target_graph.add((cloned_bnode, cloned_p, cloned_o))
    return cloned_bnode


def clone_literal(graph, node, target_graph):
    lex_val_string = str(node)
    lang = node.language
    datatype = node.datatype
    new_literal = rdflib.Literal(lex_val_string, lang, datatype)
    return new_literal


def clone_node(graph, node, target_graph, recursion=0):
    if isinstance(node, rdflib.Literal):
        new_node = clone_literal(graph, node, target_graph)
    elif isinstance(node, rdflib.BNode):
        new_node = clone_blank_node(graph, node, target_graph, recursion=recursion + 1)
    elif isinstance(node, rdflib.URIRef):
        new_node = rdflib.URIRef(str(node))
    else:
        new_node = rdflib.term.Identifier(str(node))
    return new_node



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/108
"""
from pyshacl import validate

# from DASH
shacl_file_text = """
@prefix dash: <http://datashapes.org/dash#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

dash:toString a sh:JSFunction,
        sh:SPARQLFunction ;
    rdfs:label "to string" ;
    dash:cachable true ;
    rdfs:comment "Returns a literal with datatype xsd:string that has the input value as its string. If the input value is an (URI) resource then its URI will be used." ;
    sh:jsFunctionName "dash_toString" ;
    sh:jsLibrary dash:DASHJSLibrary ;
    sh:labelTemplate "Convert {$arg} to xsd:string" ;
    sh:parameter [ a sh:Parameter ;
            sh:description "The input value." ;
            sh:name "arg" ;
            sh:nodeKind sh:IRIOrLiteral ;
            sh:path dash:arg ] ;
    sh:prefixes <http://datashapes.org/dash> ;
    sh:returnType xsd:string ;
    sh:select "SELECT (xsd:string($arg) AS ?result) WHERE { }" .
"""

def test_108():
    res = validate(shacl_file_text, advanced=True, js=True)
    conforms, graph, string = res
    assert conforms



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/79
"""
import rdflib
import pyshacl
shapes_data = """\
@prefix ex: <https://example.com#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .

ex:AAA a owl:Class.
ex:BBB a owl:Class.
ex:CCC a owl:Class.

ex:Shape a sh:NodeShape ;
  sh:targetClass ex:AAA;
  sh:not [
    sh:or (
      [sh:class ex:BBB]
      [sh:class ex:CCC]
    )
  ].
"""


def test_079():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data="""
    @prefix ex: <https://example.com#> .

    ex:aaa a ex:AAA.
    ex:aaa a ex:BBB.
    ex:aaa a ex:CCC.
    """, format='turtle')

    conforms, results_graph, results_text = pyshacl.validate(
        data_g, shacl_graph=shape_g, debug=True,
    )
    assert not conforms


if __name__ == "__main__":
    exit(test_079())



# -*- coding: utf-8 -*-

"""
https://github.com/RDFLib/pySHACL/issues/120
"""

import rdflib

from pyshacl import validate


shacl_file = r'''
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix ex: <http://example.org/> .
@prefix : <http://example.org/shape/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://example.org/shape>
  rdf:type owl:Ontology ;
  sh:declare [
      rdf:type sh:PrefixDeclaration ;
      sh:namespace "http://example.org/"^^xsd:anyURI ;
      sh:prefix "ex" ;
    ] ;
.

:exShape a sh:NodeShape ;
    sh:targetClass ex:Person ;
    sh:property [ a sh:PropertyShape;
        sh:path ( ex:i ex:name ) ;
        sh:sparql [
            sh:prefixes <http://example.org/shape> ;
            sh:message "{$this} nome does not match {?name}" ;
            sh:select """
            SELECT $this ?name
            WHERE {
                $this $PATH ?name .
                FILTER NOT EXISTS {
                    $this ex:i/ex:nome ?name .
                }
            }
        """ ] ;
    ] ;
.
'''


data_file = """
@prefix geo: <http://www.opengis.net/ont/geosparql#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix ex: <http://example.org/> .

ex:intermediate1 a ex:Intermediate ;
  ex:name "Intermediate1" ;
  ex:nome "Intermediate1" .

ex:intermediate2 a ex:Intermediate ;
  ex:name "Intermediate2" ;
  ex:nome "BadNome" .

ex:goodPerson a ex:Person ;
  ex:i ex:intermediate1 .

ex:badPerson a ex:Person ;
  ex:i ex:intermediate2 .

"""


def test_120() -> None:
    data = rdflib.Graph()
    data.parse(data=data_file, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")
    res = validate(
        data,
        shacl_graph=shapes,
        data_graph_format='turtle',
        shacl_graph_format='turtle',
        debug=True,
    )
    conforms, graph, string = res
    assert False == conforms
    assert "ex:badPerson" in string
    assert "ex:goodPerson" not in string
    assert "nome does not match Intermediate2" in string


if __name__ == "__main__":
    test_120()



from decimal import Decimal
from logging import Logger
from typing import Union

from rdflib import Literal, URIRef

from .consts import SH_datatype, SH_nodeKind, SH_optional, SH_order, SH_path
from .errors import ConstraintLoadError, ReportableRuntimeError
from .shape import Shape


class SHACLParameter(Shape):
    __slots__ = ("nodekind", "datatype", "param_order", "optional")

    def __init__(self, sg, param_node, path=None, logger: Union[Logger, None] = None):
        """
        :type sg: ShapesGraph
        :type param_node: URIRef
        """
        if path is None:
            paths = list(sg.objects(param_node, SH_path))
            if len(paths) < 1:
                path = URIRef("http://")  # todo: is this a blank path?
            elif len(paths) > 1:
                raise ConstraintLoadError(
                    "sh:parameter cannot have more than one value for sh:path.",
                    "https://www.w3.org/TR/shacl-af/#functions-example",
                )
            else:
                path = paths[0]
        super(SHACLParameter, self).__init__(sg, param_node, p=True, path=path, logger=logger)

        nodekinds = list(sg.objects(self.node, SH_nodeKind))
        if len(nodekinds) < 1:
            self.nodekind = None
        elif len(nodekinds) > 1:
            raise ConstraintLoadError(
                "sh:parameter cannot have more than one value for sh:nodeKind.",
                "https://www.w3.org/TR/shacl-af/#functions-example",
            )
        else:
            self.nodekind = nodekinds[0]

        datatypes = list(sg.objects(self.node, SH_datatype))
        if len(datatypes) < 1:
            self.datatype = None
        elif len(datatypes) > 1:
            raise ConstraintLoadError(
                "sh:parameter cannot have more than one value for sh:datatype.",
                "https://www.w3.org/TR/shacl-af/#functions-example",
            )
        else:
            self.datatype = datatypes[0]

        orders = list(sg.objects(self.node, SH_order))
        if len(orders) < 1:
            self.param_order = None
        elif len(orders) > 1:
            raise ConstraintLoadError(
                "sh:parameter cannot have more than one value for sh:order.",
                "https://www.w3.org/TR/shacl-af/#functions-example",
            )
        else:
            order = orders[0]
            if not isinstance(order, Literal):
                raise ConstraintLoadError(
                    "sh:order value must be a literal of type Decimal or Integer",
                    "https://www.w3.org/TR/shacl-af/#functions-example",
                )
            if isinstance(order.value, Decimal):
                order = order.value
            elif isinstance(order.value, int):
                order = Decimal(order.value)
            elif isinstance(order.value, float):
                order = Decimal(str(order.value))
            else:
                raise ConstraintLoadError(
                    "sh:order value must be a literal of type Decimal or Integer",
                    "https://www.w3.org/TR/shacl-af/#functions-example",
                )
            self.param_order = order
        optionals = list(sg.objects(self.node, SH_optional))
        if len(optionals) < 1:
            self.optional = False
        elif len(optionals) > 1:
            raise ConstraintLoadError(
                "sh:parameter cannot have more than one value for sh:optional.",
                "https://www.w3.org/TR/shacl-af/#functions-example",
            )
        else:
            o = optionals[0]
            if not (isinstance(o, Literal) and isinstance(o.value, bool)):
                # TODO:coverage: we don't have any tests for invalid constraints
                raise ConstraintLoadError(
                    "A sh:parameter value for sh:optional must be a valid RDF Literal of type xsd:boolean.",
                    "https://www.w3.org/TR/shacl/#constraint-components-parameters",
                )
            self.optional = bool(optionals[0])

    def __str__(self):
        name = str(self.node)
        return "<Parameter {}>".format(name)

    @property
    def localname(self):
        path = self.path()
        hash_index = path.find('#')
        if hash_index > 0:
            ending = path[hash_index + 1 :]
            return ending
        right_slash_index = path.rfind('/')
        if right_slash_index > 0:
            # TODO:coverage: No test for this case where path has a forwardslash separator
            ending = path[right_slash_index + 1 :]
            return ending
        raise ReportableRuntimeError("Cannot get a local name for {}".format(path))



import pickle

from rdflib import Graph
from rdflib.plugins.stores.memory import Memory

from pyshacl.monkey import apply_patches


apply_patches()
identifier = "http://datashapes.org/schema"
store = Memory(identifier=identifier)
with open("./schema.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier).parse(file=f)
with open("./schema.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=4)  # protocol 5 only works in python 3.8+

identifier = "http://www.w3.org/ns/shacl#"
store = Memory(identifier=identifier)
with open("./shacl.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier).parse(file=f)
with open("./shacl.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=4)  # protocol 5 only works in python 3.8+

identifier = "http://www.w3.org/ns/shacl-shacl#"
store = Memory(identifier=identifier)
with open("./shacl-shacl.ttl", "rb") as f:
    g = Graph(store=store, identifier=identifier).parse(file=f)
with open("./shacl-shacl.pickle", "wb") as f:
    pickle.dump((store, identifier), f, protocol=4)  # protocol 5 only works in python 3.8+



# -*- coding: utf-8 -*-


class ReportableRuntimeError(RuntimeError):
    def __init__(self, message):
        self.message = message

    @property
    def args(self):
        return [self.message]

    def __str__(self):
        return str(self.message)

    def __repr__(self):
        return "ReportableRuntimeError: {}".format(self.__str__())


class ShapeLoadError(ReportableRuntimeError):
    def __init__(self, message, link):
        self.message = message
        self.link = link

    @property
    def args(self):
        return [self.message, self.link]

    def __str__(self):
        return "{}\nFor reference, see {}".format(str(self.message), str(self.link))

    def __repr__(self):
        return "ShapeLoadError: {}".format(self.__str__())


class RuleLoadError(ReportableRuntimeError):
    def __init__(self, message, link):
        self.message = message
        self.link = link

    @property
    def args(self):
        return [self.message, self.link]

    def __str__(self):
        return "{}\nFor reference, see {}".format(str(self.message), str(self.link))

    def __repr__(self):
        return "RuleLoadError: {}".format(self.__str__())


class ValidationFailure(ReportableRuntimeError):
    """
    https://www.w3.org/TR/shacl/#failures
    Validation and conformance checking can result in a failure. For example, a particular SHACL processor might allow recursive shapes but report a failure if it detects a loop within the data. Failures can also be reported due to resource exhaustion. Failures are signalled through implementation-specific channels.
    """

    def __init__(self, message):
        self.message = message

    @property
    def args(self):
        return [self.message]

    def __str__(self):
        return str(self.message)

    def __repr__(self):
        return "ValidationFailure: {}".format(self.__str__())


class ValidationWarning(RuntimeWarning):
    def __init__(self, message, link):
        super(ValidationWarning, self).__init__()
        self.message = message
        self.link = link

    @property
    def args(self):
        return [self.message, self.link]

    def __str__(self):
        return "{}\nFor reference, see {}".format(str(self.message), str(self.link))

    def __repr__(self):
        return "{}: {}".format(str(self.__class__), self.__str__())


class ShapeRecursionWarning(ValidationWarning):
    def __init__(self, evaluation_path):
        length = len(evaluation_path)
        r_string = "->".join(str(e) for e in evaluation_path)
        message = (
            "Warning, A Recursive Shape was detected executing a recursive validation sequence "
            "{} levels deep. Backing out.\n{}".format(length, r_string)
        )
        link = "https://www.w3.org/TR/shacl/#shapes-recursion"
        super(ShapeRecursionWarning, self).__init__(message, link)


class ConstraintLoadError(ReportableRuntimeError):
    def __init__(self, message, link):
        self.message = message
        self.link = link

    @property
    def args(self):
        return [self.message, self.link]

    def __str__(self):
        return "{}\nFor reference, see {}".format(str(self.message), str(self.link))

    def __repr__(self):
        return "ConstraintLoadError: {}".format(self.__str__())


class ConstraintLoadWarning(RuntimeWarning):
    def __init__(self, message, link):
        super(ConstraintLoadWarning, self).__init__()
        self.message = message
        self.link = link

    @property
    def args(self):
        return [self.message, self.link]

    def __str__(self):
        return "{}\nFor reference, see {}".format(str(self.message), str(self.link))

    def __repr__(self):
        return "ConstraintLoadWarning: {}".format(self.__str__())



# -*- coding: utf-8 -*-
#
from typing import Union

from rdflib import RDF, RDFS, ConjunctiveGraph, Dataset, Graph, Namespace
from rdflib.term import Node


RDFS_Resource = RDFS.Resource
RDF_first = RDF.first
SH = Namespace('http://www.w3.org/ns/shacl#')

ConjunctiveLike = Union[ConjunctiveGraph, Dataset]
GraphLike = Union[ConjunctiveLike, Graph]
RDFNode = Node



#
#
import typing

from typing import Dict, List
from warnings import warn

from rdflib import URIRef

from pyshacl.consts import SH_JSTargetType
from pyshacl.errors import ShapeLoadError
from pyshacl.target import BoundSHACLTargetType, SHACLTargetType

from .js_executable import JSExecutable


if typing.TYPE_CHECKING:
    from pyshacl.pytypes import GraphLike
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph


class JSTarget(JSExecutable):
    def __init__(self, shapes_graph: 'ShapesGraph', exe_node):
        super(JSTarget, self).__init__(shapes_graph, exe_node)

    def find_targets(self, data_graph):
        results = self.execute(data_graph, {}, mode='target')
        return [u for u in results['_result'] if isinstance(u, URIRef)]


class BoundJSTargetType(BoundSHACLTargetType):
    __slots__ = ('params_kv',)

    def __init__(self, target_type: 'JSTargetType', target_declaration, shape: 'Shape', params_kv):
        super(BoundJSTargetType, self).__init__(target_type, target_declaration, shape)
        self.params_kv = params_kv  # type: dict

    @classmethod
    def constraint_parameters(cls):
        return []

    @classmethod
    def constraint_name(cls):
        return "JSTargetType"

    @classmethod
    def shacl_constraint_class(cls):
        return SH_JSTargetType

    def evaluate(self, target_graph: 'GraphLike', focus_value_nodes: Dict, _evaluation_path: List):
        raise NotImplementedError()

    def find_targets(self, data_graph):
        results = self.target_type.js_exe.execute(data_graph, self.params_kv, mode='target')
        return [u for u in results['_result'] if isinstance(u, URIRef)]


class JSTargetType(SHACLTargetType):
    __slots__ = ('js_exe',)

    def __init__(self, tt_node, sg: 'ShapesGraph'):
        super(JSTargetType, self).__init__(tt_node, sg)
        self.js_exe = JSExecutable(sg, tt_node)

    def check_params(self, target_declaration):
        param_kv = {}
        for p in self.parameters:
            path = p.path()
            name = p.localname
            vals = set(self.sg.objects(target_declaration, path))
            if len(vals) < 1:
                if p.optional:
                    continue
                raise ShapeLoadError(
                    "sh:target does not have a value for {}".format(name),
                    "https://www.w3.org/TR/shacl-js/#JSTargetType",
                )
            if len(vals) > 1:
                warn(Warning("Found more than one value for {} on sh:target. Using just first one.".format(name)))
            param_kv[name] = next(iter(vals))
        return param_kv

    def bind(self, shape, target_declaration):
        param_vals = self.check_params(target_declaration)
        return BoundJSTargetType(self, target_declaration, shape, param_vals)



"""\
A cool test that combines a bunch of SHACL-AF features, including:
SHACL Functions (implemented as SPARQL functions)
SHACL Rules
Node Expressions
Expression Constraint
"""

from pyshacl import validate
from rdflib import Graph

shacl_file = '''\
# prefix: ex

@prefix ex: <http://datashapes.org/shasf/tests/expression/advanced.test.shacl#> .
@prefix exOnt: <http://datashapes.org/shasf/tests/expression/advanced.test.ont#> .
@prefix exData: <http://datashapes.org/shasf/tests/expression/advanced.test.data#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://datashapes.org/shasf/tests/expression/advanced.test.shacl>
  rdf:type owl:Ontology ;
  rdfs:label "Test of advanced features" ;
.

ex:concat
    a sh:SPARQLFunction ;
    rdfs:comment "Concatenates strings $op1 and $op2." ;
    sh:parameter [
        sh:path ex:op1 ;
        sh:datatype xsd:string ;
        sh:description "The first string" ;
    ] ;
    sh:parameter [
        sh:path ex:op2 ;
        sh:datatype xsd:string ;
        sh:description "The second string" ;
    ] ;
    sh:returnType xsd:string ;
    sh:select """
        SELECT ?result
        WHERE {
          BIND(CONCAT(STR(?op1),STR(?op2)) AS ?result) .
        }
        """ .

ex:strlen
    a sh:SPARQLFunction ;
    rdfs:comment "Returns length of the given string." ;
    sh:parameter [
        sh:path ex:op1 ;
        sh:datatype xsd:string ;
        sh:description "The string" ;
    ] ;
    sh:returnType xsd:integer ;
    sh:select """
        SELECT ?result
        WHERE {
          BIND(STRLEN(?op1) AS ?result) .
        }
        """ .

ex:lessThan
    a sh:SPARQLFunction ;
    rdfs:comment "Returns True if op1 < op2." ;
    sh:parameter [
        sh:path ex:op1 ;
        sh:datatype xsd:integer ;
        sh:description "The first int" ;
    ] ;
    sh:parameter [
        sh:path ex:op2 ;
        sh:datatype xsd:integer ;
        sh:description "The second int" ;
    ] ;
    sh:returnType xsd:boolean ;
    sh:select """
        SELECT ?result
        WHERE {
          BIND(IF(?op1 < ?op2, true, false) AS ?result) .
        }
        """ .

ex:PersonExpressionShape
    a sh:NodeShape ;
    sh:targetClass exOnt:Person ;
    sh:expression [
        sh:message "Person's firstName and lastName together should be less than 35 chars long." ;
        ex:lessThan (
            [ ex:strlen (
                [ ex:concat ( [ sh:path exOnt:firstName] [ sh:path exOnt:lastName ] ) ] )
            ]
            35 );
    ] .

ex:PersonRuleShape
	a sh:NodeShape ;
	sh:targetClass exOnt:Administrator ;
	sh:message "An administrator is a person too." ;
	sh:rule [
		a sh:TripleRule ;
		sh:subject sh:this ;
		sh:predicate rdf:type ;
		sh:object exOnt:Person ;
	] .
'''

data_graph = '''
# prefix: ex

@prefix ex: <http://datashapes.org/shasf/tests/expression/advanced.test.data#> .
@prefix exOnt: <http://datashapes.org/shasf/tests/expression/advanced.test.ont#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:Kate
  rdf:type exOnt:Person ;
  exOnt:firstName "Kate" ;
  exOnt:lastName "Jones" ;
.

ex:Jenny
  rdf:type exOnt:Administrator ;
  exOnt:firstName "Jennifer" ;
  exOnt:lastName "Wolfeschlegelsteinhausenbergerdorff" ;
.
'''

if __name__ == "__main__":
    d = Graph().parse(data=data_graph, format="turtle")
    s = Graph().parse(data=shacl_file, format="turtle")
    conforms, report, message = validate(d, shacl_graph=s, advanced=True, debug=False)
    print(message)



from pyshacl import validate
from io import BytesIO

"""
https://github.com/RDFLib/pySHACL/issues/62
"""

data = b"""\
{
  "@context": {
    "ex": "http://example.com/ex#",
    "exOnt": "http://example.com/exOnt#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "xsd": "http://www.w3.org/2001/XMLSchema#"
  },
  "@graph": [
    {
      "@id": "ex:Pet1",
      "@type": "exOnt:Lizard",
      "exOnt:nLegs": 4,
      "rdf:label": "Sebastian"
    },
    {
      "@id": "ex:Human1",
      "@type": "exOnt:Human",
      "exOnt:hasPet": {
        "@id": "ex:Pet1"
      },
      "exOnt:nLegs": 2,
      "rdf:label": "Amy"
    }
  ]
}
"""

ont_data = b"""\
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exOnt> a owl:Ontology ;
    rdfs:label "An example extra-ontology file."@en .

exOnt:Animal a rdfs:Class ;
    rdfs:comment "The parent class for Humans and Pets"@en ;
    rdfs:subClassOf owl:Thing .

exOnt:Human a rdfs:Class ;
    rdfs:comment "A Human being"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:Pet a rdfs:Class ;
    rdfs:comment "An animal owned by a human"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:hasPet a rdf:Property ;
    rdfs:domain exOnt:Human ;
    rdfs:range exOnt:Pet .

exOnt:nlegs a rdf:Property ;
    rdfs:domain exOnt:Animal ;
    rdfs:range exOnt:integer .

exOnt:Lizard a rdfs:Class ;
    rdfs:subClassOf exOnt:Pet .
"""

shacl_data = b"""\
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exShape: <http://example.com/exShape#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exShape> a owl:Ontology ;
    rdfs:label "Example Shapes File"@en .

exShape:HumanShape a sh:NodeShape ;
    sh:property [
        sh:class exOnt:Pet ;
        sh:path exOnt:hasPet ;
    ] ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 2 ;
        sh:minInclusive 2 ;
    ] ;
    sh:targetClass exOnt:Human .

exShape:AnimalShape a sh:NodeShape ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 4 ;
        sh:minInclusive 1 ;
    ] ;
    sh:targetClass exOnt:Animal .
"""

def test_062():
    data_fp = BytesIO(data)
    shacl_fp = BytesIO(shacl_data)
    ont_fp = BytesIO(ont_data)
    try:
        conforms, g, s = validate(data_fp, shacl_graph=shacl_fp, ont_graph=ont_fp, data_graph_format="json-ld",
                                  shacl_graph_format="turtle", ont_graph_format="turtle", abort_on_first=False,
                                  meta_shacl=False, debug=True, advanced=True)
    except Exception as e:
        print(e)
        raise
    assert conforms

if __name__ == "__main__":
    test_062()



#
#
from pyshacl import validate
from pyshacl.rdfutil.load import load_from_source

data_graph = load_from_source("./test/issues/test_029/simpleData.ttl")
shacl_graph = load_from_source("./test/issues/test_029/simpleOnto.ttl")


def test_029():
    res = validate(data_graph, shacl_graph=shacl_graph, ont_graph=shacl_graph,
                   inference='none', debug=True)
    conforms, graph, string = res
    assert not conforms

if __name__ == "__main__":
    test_029()




#
#
import typing

from pyshacl.consts import SH
from pyshacl.errors import ReportableRuntimeError
from pyshacl.rules.shacl_rule import SHACLRule

from .js_executable import JSExecutable


if typing.TYPE_CHECKING:
    from pyshacl.pytypes import GraphLike
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph

SH_JSRule = SH.JSRule


class JSRule(SHACLRule):
    __slots__ = ('js_exe',)

    def __init__(self, shape: 'Shape', rule_node, **kwargs):
        super(JSRule, self).__init__(shape, rule_node, **kwargs)
        shapes_graph = shape.sg  # type: ShapesGraph
        self.js_exe = JSExecutable(shapes_graph, rule_node)

    def apply(self, data_graph: 'GraphLike') -> int:
        focus_nodes = self.shape.focus_nodes(data_graph)  # uses target nodes to find focus nodes
        all_added = 0
        iterate_limit = 100
        while True:
            if iterate_limit < 1:
                raise ReportableRuntimeError("Local rule iteration exceeded iteration limit of 100.")
            iterate_limit -= 1
            added = 0
            applicable_nodes = self.filter_conditions(focus_nodes, data_graph)
            sets_to_add = []
            for a in applicable_nodes:
                args_map = {"this": a}
                results = self.js_exe.execute(data_graph, args_map, mode="construct")
                triples = results['_result']
                this_added = False
                if triples is not None and isinstance(triples, (list, tuple)):
                    set_to_add = set()
                    for t in triples:
                        s, p, o = tr = t[:3]
                        if not this_added and tr not in data_graph:
                            this_added = True
                        set_to_add.add(tr)
                    sets_to_add.append(set_to_add)
                if this_added:
                    added += 1
            if added > 0:
                all_added += added
                for s in sets_to_add:
                    for t in s:
                        data_graph.add(t)
                if self.iterate:
                    continue  # Jump up to iterate
                else:
                    break  # Don't iterate
            break
        return all_added



# -*- coding: utf-8 -*-
#
"""
https://github.com/RDFLib/pySHACL/issues/79
"""
import rdflib
import pyshacl
shapes_data = """\
@prefix ex: <https://example.com#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

ex:ParentTypeRestriction
  a sh:NodeShape ;
  sh:targetClass ex:Cls1 ;
  sh:property [
    sh:path ex:parent ;
    sh:minCount 1 ;
    sh:maxCount 1 ;
    sh:property [
      sh:path ( rdf:type [ sh:zeroOrMorePath rdfs:subClassOf ] ) ;
      sh:minCount 1 ;
      sh:qualifiedMinCount 1 ;
      sh:qualifiedMaxCount 1 ;
      sh:qualifiedValueShape [
        sh:or (
          [ sh:hasValue ex:P_cls1 ]
          [ sh:hasValue ex:P_cls2 ]
          [ sh:hasValue ex:P_cls3 ]
        )
      ] ;
      sh:not [ sh:in ( ex:P_clsInvalid_a ex:P_clsInvalid_b ) ] ;
    ] ;
    sh:message "parent type for ex:Cls1 is not in x:P_cls{1, 2, 3}" ;
  ] ;
.
"""
# rdf:type/rdfs:subClassOf*
target_data = """\
@prefix ex: <https://example.com#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .

ex:Child1
  a ex:Cls1 ;
  ex:parent ex:Parent1 ;
.
ex:Parent1
  a ex:P_cls1, ex:P_cls4 ;
.
"""


def test_w3_list1():
    shape_g = rdflib.Graph().parse(data=shapes_data, format='turtle')
    data_g = rdflib.Graph().parse(data=target_data, format='turtle')

    conforms, results_graph, results_text = pyshacl.validate(
        data_g, shacl_graph=shape_g, debug=False,
    )
    print(results_text)
    assert conforms


if __name__ == "__main__":
    exit(test_w3_list1())



# -*- coding: utf-8 -*-
"""
https://www.w3.org/TR/shacl/#sparql-constraint-components
"""
import typing

from typing import Dict, List, Tuple, Type, Union

import rdflib

from pyshacl.constraints.constraint_component import ConstraintComponent, CustomConstraintComponent
from pyshacl.consts import SH, RDF_type, SH_ask, SH_ConstraintComponent, SH_message, SH_select
from pyshacl.errors import ConstraintLoadError, ValidationFailure
from pyshacl.helper import get_query_helper_cls
from pyshacl.pytypes import GraphLike


if typing.TYPE_CHECKING:
    from pyshacl.shape import Shape
    from pyshacl.shapes_graph import ShapesGraph


SH_SPARQLSelectValidator = SH.SPARQLSelectValidator
SH_SPARQLAskValidator = SH.SPARQLAskValidator


class BoundShapeValidatorComponent(ConstraintComponent):

    shacl_constraint_component = SH_ConstraintComponent

    def __init__(self, constraint, shape: 'Shape', validator):
        """
        Create a new custom constraint, by applying a ConstraintComponent and a Validator to a Shape
        :param constraint: The source ConstraintComponent, this is needed to bind the parameters in the query_helper
        :type constraint: SPARQLConstraintComponent
        :param shape:
        :type shape: Shape
        :param validator:
        :type validator: AskConstraintValidator | SelectConstraintValidator
        """
        super(BoundShapeValidatorComponent, self).__init__(shape)
        self.constraint = constraint
        self.validator = validator
        params = constraint.parameters
        SPARQLQueryHelper = get_query_helper_cls()
        self.query_helper = SPARQLQueryHelper(
            self.shape, validator.node, validator.query_text, params, messages=validator.messages
        )
        # Setting self.shape into QueryHelper automatically applies query_helper.bind_params and bind_messages
        self.query_helper.collect_prefixes()

    @classmethod
    def constraint_parameters(cls):
        # TODO:coverage: this is never used for this constraint?
        return []

    @classmethod
    def constraint_name(cls):
        return "ConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[rdflib.Literal]:
        return [rdflib.Literal("Parameterised SHACL Query generated constraint validation reports.")]

    def evaluate(self, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List):
        """
        :type focus_value_nodes: dict
        :type target_graph: rdflib.Graph
        """
        reports = []
        non_conformant = False
        extra_messages = self.constraint.messages or []
        rept_kwargs = {
            # TODO, determine if we need sourceConstraint here
            #  'source_constraint': self.validator.node,
            'constraint_component': self.constraint.node,
            'extra_messages': extra_messages,
        }
        for f, value_nodes in focus_value_nodes.items():
            # we don't use value_nodes in the sparql constraint
            # All queries are done on the corresponding focus node.
            try:
                violations = self.validator.validate(f, value_nodes, target_graph, self.query_helper)
            except ValidationFailure as e:
                raise e
            for val, vio in violations:
                non_conformant = True
                msg_args_map = self.query_helper.param_bind_map.copy()
                msg_args_map.update({"this": f, "value": val})
                if self.shape.is_property_shape:
                    msg_args_map['path'] = self.shape.path()
                self.query_helper.bind_messages(msg_args_map)
                bound_messages = self.query_helper.bound_messages
                # The DASH test suite likes _no_ value entry in the report if we're on a Property Shape.
                report_val = val if not self.shape.is_property_shape else None
                if isinstance(vio, bool):
                    if vio is False:  # ASKValidator Result
                        new_kwargs = rept_kwargs.copy()
                        new_kwargs['extra_messages'].extend(bound_messages)
                        rept = self.make_v_result(target_graph, f, value_node=report_val, **new_kwargs)
                    else:  # SELECTValidator Failure
                        raise ValidationFailure("Validation Failure generated by SPARQLConstraint.")
                elif isinstance(vio, tuple):
                    t, p, v = vio
                    new_msg_args_map = msg_args_map.copy()
                    if v is None:
                        v = report_val
                    else:
                        new_msg_args_map['value'] = v
                    if p is not None:
                        new_msg_args_map['path'] = p
                    if t is not None:
                        new_msg_args_map['this'] = t
                    self.query_helper.bind_messages(new_msg_args_map)
                    new_bound_msgs = self.query_helper.bound_messages
                    new_kwargs = rept_kwargs.copy()
                    new_kwargs['extra_messages'].extend(new_bound_msgs)
                    rept = self.make_v_result(target_graph, t or f, value_node=v, result_path=p, **new_kwargs)
                else:
                    new_kwargs = rept_kwargs.copy()
                    new_kwargs['extra_messages'].extend(bound_messages)
                    rept = self.make_v_result(target_graph, f, value_node=report_val, **new_kwargs)
                reports.append(rept)
        return (not non_conformant), reports


class SPARQLConstraintComponentValidator(object):
    validator_cache: Dict[Tuple[int, str], Union['SelectConstraintValidator', 'AskConstraintValidator']] = {}

    def __new__(cls, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        cache_key = (id(shacl_graph.graph), str(node))
        found_in_cache = cls.validator_cache.get(cache_key, False)
        if found_in_cache:
            return found_in_cache
        sg = shacl_graph.graph
        type_vals = set(sg.objects(node, RDF_type))
        validator_type: Union[Type[SelectConstraintValidator], Type[AskConstraintValidator], None] = None
        if len(type_vals) > 0:
            if SH_SPARQLSelectValidator in type_vals:
                validator_type = SelectConstraintValidator
            elif SH_SPARQLAskValidator in type_vals:
                validator_type = AskConstraintValidator
        if not validator_type:
            sel_nodes = set(sg.objects(node, SH_select))
            if len(sel_nodes) > 0:
                # TODO:coverage: No test for this case
                validator_type = SelectConstraintValidator
        if not validator_type:
            ask_nodes = set(sg.objects(node, SH_ask))
            if len(ask_nodes) > 0:
                validator_type = AskConstraintValidator

        if not validator_type:
            # TODO:coverage: No test for this case
            raise ConstraintLoadError(
                "Validator must be of type sh:SPARQLSelectValidator or sh:SPARQLAskValidator and must have either a sh:select or a sh:ask predicate.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )
        validator = validator_type(shacl_graph, node, *args, **kwargs)
        cls.validator_cache[cache_key] = validator
        return validator

    def apply_to_shape_via_constraint(self, constraint, shape, **kwargs) -> BoundShapeValidatorComponent:
        """
        Create a new Custom Constraint (BoundShapeValidatorComponent)
        :param constraint:
        :type constraint: SPARQLConstraintComponent
        :param shape:
        :type shape: pyshacl.shape.Shape
        :param kwargs:
        :return:
        """
        must_be_ask_val = kwargs.pop('must_be_ask_val', False)
        if must_be_ask_val and not (isinstance(self, AskConstraintValidator)):
            # TODO:coverage: No test for this case, do we need to test this?
            raise ConstraintLoadError(
                "Validator not for NodeShape or a PropertyShape must be of type SPARQLAskValidator.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )
        must_be_select_val = kwargs.pop('must_be_select_val', False)
        if must_be_select_val and not (isinstance(self, SelectConstraintValidator)):
            # TODO:coverage: No test for this case, do we need to test this?
            raise ConstraintLoadError(
                "Validator for a NodeShape or a PropertyShape must be of type SPARQLSelectValidator.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )

        return BoundShapeValidatorComponent(constraint, shape, self)

    def __init__(self, shacl_graph: 'ShapesGraph', node, **kwargs):
        initialised = getattr(self, 'initialised', False)
        if initialised:
            return
        self.shacl_graph = shacl_graph
        self.node = node
        sg = shacl_graph.graph
        message_nodes = set(sg.objects(node, SH_message))
        for m in message_nodes:
            if not (isinstance(m, rdflib.Literal) and isinstance(m.value, str)):
                # TODO:coverage: No test for when SPARQL-based constraint is RDF Literal is is not of type string
                raise ConstraintLoadError(
                    "Validator sh:message must be an RDF Literal of type xsd:string.",
                    "https://www.w3.org/TR/shacl/#ConstraintComponent",
                )
        self.messages = message_nodes
        self.initialised = True

    def make_messages(self, params_map=None):
        if params_map is None:
            return self.messages
        ret_msgs = []
        for m in self.messages:
            this_m = m.value[:]
            for a, v in params_map.items():
                replace_me = "{$" + str(a) + "}"
                if isinstance(v, rdflib.Literal):
                    v = v.value
                this_m = this_m.replace(replace_me, str(v))
            ret_msgs.append(rdflib.Literal(this_m))
        return ret_msgs


class AskConstraintValidator(SPARQLConstraintComponentValidator):
    def __new__(cls, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        return object.__new__(cls)

    def __init__(self, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        super(AskConstraintValidator, self).__init__(shacl_graph, node, **kwargs)
        g = shacl_graph.graph
        ask_vals = set(g.objects(node, SH_ask))
        if len(ask_vals) < 1 or len(ask_vals) > 1:
            # TODO:coverage: No test for this case
            raise ConstraintLoadError(
                "AskValidator must have exactly one value for sh:ask.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )
        ask_val = next(iter(ask_vals))
        if not (isinstance(ask_val, rdflib.Literal) and isinstance(ask_val.value, str)):
            # TODO:coverage: No test for this case
            raise ConstraintLoadError(
                "AskValidator sh:ask must be an RDF Literal of type xsd:string.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )
        self.query_text = ask_val.value

    def validate(self, focus, value_nodes, target_graph, query_helper=None, new_bind_vals=None):
        """

        :param focus:
        :param value_nodes:
        :param query_helper:
        :param target_graph:
        :type target_graph: rdflib.Graph
        :param new_bind_vals:
        :return:
        """
        param_bind_vals = query_helper.param_bind_map if query_helper else {}
        new_bind_vals = new_bind_vals or {}
        bind_vals = param_bind_vals.copy()
        bind_vals.update(new_bind_vals)
        violations = set()
        for v in value_nodes:
            if query_helper is None:
                # TODO:coverage: No test for this case when query_helper is None
                init_binds = {}
                sparql_text = self.query_text
            else:
                init_binds, sparql_text = query_helper.pre_bind_variables(
                    focus, valuenode=v, extravars=bind_vals.keys()
                )
                sparql_text = query_helper.apply_prefixes(sparql_text)
                init_binds.update(bind_vals)
            try:
                result = target_graph.query(sparql_text, initBindings=init_binds)
                answer = result.askAnswer
            except (KeyError, AttributeError):
                # TODO:coverage: Can this ever actually happen?
                raise ValidationFailure("ASK Query did not return an askAnswer.")
            if answer is False:
                violations.add((v, False))
        return violations


class SelectConstraintValidator(SPARQLConstraintComponentValidator):
    def __new__(cls, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        return object.__new__(cls)

    def __init__(self, shacl_graph: 'ShapesGraph', node, *args, **kwargs):
        super(SelectConstraintValidator, self).__init__(shacl_graph, node, **kwargs)
        g = shacl_graph.graph
        select_vals = set(g.objects(node, SH_select))
        if len(select_vals) < 1 or len(select_vals) > 1:
            # TODO:coverage: No test for this case, do we need to test this?
            raise ConstraintLoadError(
                "SelectValidator must have exactly one value for sh:select.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )
        select_val = next(iter(select_vals))
        if not (isinstance(select_val, rdflib.Literal) and isinstance(select_val.value, str)):
            # TODO:coverage: No test for the case when sh:select is not a literal of type string
            raise ConstraintLoadError(
                "SelectValidator sh:select must be an RDF Literal of type xsd:string.",
                "https://www.w3.org/TR/shacl/#ConstraintComponent",
            )
        self.query_text = select_val.value

    def validate(self, focus, value_nodes, target_graph, query_helper=None, new_bind_vals=None):
        """

        :param focus:
        :param value_nodes:
        :param query_helper:
        :param target_graph:
        :type target_graph: rdflib.Graph
        :param new_bind_vals:
        :return:
        """
        param_bind_vals = query_helper.param_bind_map if query_helper else {}
        new_bind_vals = new_bind_vals or {}
        bind_vals = param_bind_vals.copy()
        bind_vals.update(new_bind_vals)
        violations = set()
        for v in value_nodes:
            if query_helper is None:
                # TODO:coverage: No test for this case when query_helper is None
                init_binds = {}
                sparql_text = self.query_text
            else:
                init_binds, sparql_text = query_helper.pre_bind_variables(
                    focus, valuenode=v, extravars=bind_vals.keys()
                )
                sparql_text = query_helper.apply_prefixes(sparql_text)
                init_binds.update(bind_vals)
            results = target_graph.query(sparql_text, initBindings=init_binds)
            if not results or len(results.bindings) < 1:
                continue
            for r in results:
                try:
                    p = r['path']
                except KeyError:
                    p = None
                try:
                    v2 = r['value']
                except KeyError:
                    v2 = None
                try:
                    t = r['this']
                except KeyError:
                    # TODO:coverage: No test for when result has no 'this' key
                    t = None
                if p or v2 or t:
                    violations.add((v, (t, p, v2)))
                else:
                    # TODO:coverage: No test for generic failure, when
                    #  'path' and 'value' and 'this' are not returned.
                    #  here 'failure' must exist
                    try:
                        f = r['failure']
                        if f is True or (isinstance(f, rdflib.Literal) and f.value):
                            violations.add((v, True))
                    except KeyError:
                        pass
        return violations


class SPARQLConstraintComponent(CustomConstraintComponent):
    """
    SPARQL-based constraints provide a lot of flexibility but may be hard to understand for some people or lead to repetition. This section introduces SPARQL-based constraint components as a way to abstract the complexity of SPARQL and to declare high-level reusable components similar to the Core constraint components. Such constraint components can be declared using the SHACL RDF vocabulary and thus shared and reused.
    Link:
    https://www.w3.org/TR/shacl/#sparql-constraint-components
    """

    __slots__: Tuple = tuple()

    def __new__(cls, shacl_graph, node, parameters, validators, node_validators, property_validators):
        return super(SPARQLConstraintComponent, cls).__new__(
            cls, shacl_graph, node, parameters, validators, node_validators, property_validators
        )

    @property
    def messages(self):
        # TODO: allow messages at this SPARQLConstraintComponent level
        return []

    def make_validator_for_shape(self, shape: 'Shape'):
        """
        :param shape:
        :type shape: Shape
        :return:
        """
        val_count = len(self.validators)
        node_val_count = len(self.node_validators)
        prop_val_count = len(self.property_validators)
        must_be_select_val = False
        must_be_ask_val = False
        if shape.is_property_shape and prop_val_count > 0:
            validator_node = next(iter(self.property_validators))
            must_be_select_val = True
        elif (not shape.is_property_shape) and node_val_count > 0:
            validator_node = next(iter(self.node_validators))
            must_be_select_val = True
        elif val_count > 0:
            validator_node = next(iter(self.validators))
            must_be_ask_val = True
        else:
            raise ConstraintLoadError(
                "Cannot select a validator to use, according to the rules.",
                "https://www.w3.org/TR/shacl/#constraint-components-validators",
            )

        validator = SPARQLConstraintComponentValidator(self.sg, validator_node)
        applied_validator = validator.apply_to_shape_via_constraint(
            self, shape, must_be_ask_val=must_be_ask_val, must_be_select_val=must_be_select_val
        )
        return applied_validator



#
#
import typing

from rdflib.plugins.sparql.operators import register_custom_function, unregister_custom_function
from rdflib.plugins.sparql.sparql import SPARQLError

from pyshacl.errors import ReportableRuntimeError
from pyshacl.functions.shacl_function import SHACLFunction

from .js_executable import JSExecutable


if typing.TYPE_CHECKING:
    from pyshacl.shapes_graph import ShapesGraph


class JSFunction(SHACLFunction):
    __slots__ = ('js_exe',)

    def __init__(self, fn_node, shapes_graph: 'ShapesGraph'):
        super(JSFunction, self).__init__(fn_node, shapes_graph)
        self.js_exe = JSExecutable(shapes_graph, fn_node)

    def execute(self, g, *args):
        params = self.get_params_in_order()
        if len(args) != len(params):
            raise ReportableRuntimeError("Got incorrect number of arguments for JSFunction {}.".format(self.node))
        args_map = {}
        for i, p in enumerate(params):
            arg = args[i]
            ln = p.localname
            if arg is None and p.optional is False:
                raise ReportableRuntimeError("Got NoneType for Non-optional argument {}.".format(ln))
            args_map[ln] = arg
        results = self.js_exe.execute(g, args_map, mode="function", return_type=self.rtype)
        res = results['_result']
        return res

    def execute_from_sparql(self, e, ctx):
        if not e.expr:
            raise SPARQLError("Nothing given to SPARQLFunction.")
        params = self.get_params_in_order()
        num_params = len(params)
        if len(e.expr) > num_params:
            raise SPARQLError("Too many parameters passed to SPARQLFunction.")
        elif len(e.expr) < num_params:
            raise SPARQLError("Too few parameters passed to SPARQLFunction.")
        args_map = {str(var): val for var, val in ctx.ctx.initBindings.items()}
        args_map.update({str(var): val for var, val in ctx.ctx.bindings.items()})
        g = ctx.ctx.graph
        for i, var in enumerate(e.expr):
            var_val = ctx[var]
            param_name = params[i].localname
            args_map[param_name] = var_val
        results = self.js_exe.execute(g, args_map, mode="function", return_type=self.rtype)
        res = results['_result']
        return res

    def apply(self, g):
        super(JSFunction, self).apply(g)
        register_custom_function(self.node, self.execute_from_sparql, True, True)

    def unapply(self, g):
        super(JSFunction, self).unapply(g)
        unregister_custom_function(self.node, self.execute_from_sparql)



import rdflib
from pyshacl import validate
from pyshacl.errors import ConstraintLoadError

"""
https://github.com/RDFLib/pySHACL/issues/58
"""
data = """
@prefix asdf: <http://example.org/asdf/> .
@prefix ex: <http://example.org/> .

asdf:e2e a ex:termA ;
    ex:child asdf:23e .

asdf:23e a ex:termB .
"""

shaclData = """
@prefix ex: <http://example.org/> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .

ex:termShape a sh:NodeShape ;
    sh:ignoredProperties ( rdf:type ) ;
    sh:targetClass ex:termB .
"""



def test_058():
    dataGraph = rdflib.Graph().parse(data=data, format='ttl')
    shaclGraph = rdflib.Graph().parse(data=shaclData, format='ttl')
    exc = None
    try:
        conforms, g, s = validate(dataGraph, shacl_graph=shaclGraph, abort_on_first=False, meta_shacl=False, debug=False, advanced=True)
    except Exception as e:
        assert isinstance(e, ConstraintLoadError)
        exc = e
    assert exc is not None
    assert "ignoredProperties" in str(exc)



if __name__ == "__main__":
    test_058()
